from pytest import mark, raises

from apischema.validation.dependencies import find_all_dependencies, find_dependencies


def a_equal_b(param):
    assert param.a == param.b


@mark.parametrize("func, deps", [(a_equal_b, {"a", "b"}), (int, set())])
def test_find_dependencies(func, deps):
    assert find_dependencies(func) == deps


def test_no_parameter():
    with raises(TypeError):
        find_dependencies(lambda: None)


def test_find_end_dependencies():
    class Test:
        class_var = ""

        def __init__(self):
            self.a = 0
            self.b = {}

        def pseudo_validate(self):
            if self.a not in self.method(0):
                yield self.class_var

        def method(self, arg):
            res = list(self.c)
            if len(res) < arg:
                return self.method(arg - 1)

        @property
        def c(self):
            return self.b.values()

    assert find_all_dependencies(Test, Test.pseudo_validate) == {"a", "b", "class_var"}



from enum import Enum
from typing import (  # type: ignore
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from apischema.conversion.utils import Conversions
from apischema.json_schema.generation.visitor import SchemaVisitor
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.types import AnyType
from apischema.typing import Annotated
from apischema.utils import is_hashable

Refs = Dict[str, Tuple[AnyType, int]]


class Recursive(Exception):
    pass


class RefsExtractor(SchemaVisitor):
    def __init__(self, conversions: Optional[Conversions], refs: Refs):
        super().__init__(conversions)
        self.refs = refs

    def _incr_ref(self, ref: Optional[str], cls: AnyType) -> bool:
        if ref is not None:
            ref_cls, count = self.refs.get(ref, (cls, 0))
            if ref_cls != cls:
                raise ValueError(
                    f"Types {cls} and {self.refs[ref]} share same reference '{ref}'"
                )
            self.refs[ref] = (ref_cls, count + 1)
            return count > 0
        return False

    def _annotated(self, cls: AnyType, annotations: Sequence[Any], _):
        for annotation in reversed(annotations):
            if isinstance(annotation, schema_ref):
                annotation.check_type(cls)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
                annotated = Annotated[(cls, *filter(is_hashable, annotations))]  # type: ignore # noqa E501
                if self._incr_ref(ref, annotated):
                    return
        return self.visit(cls)

    def any(self, _):
        pass

    def collection(self, cls: Type[Iterable], value_type: AnyType, _):
        return self.visit(value_type)

    def dataclass(self, cls: Type, _):
        (
            fields,
            merged_fields,
            pattern_fields,
            additional_field,
        ) = self._dataclass_fields(cls)
        for field in fields:
            self._field_visit(field, _)
        for _, field in merged_fields:
            self._field_visit(field, _)
        for _, field in pattern_fields:
            self._field_visit(field, _)
        if additional_field:
            self._field_visit(additional_field, _)

    def enum(self, cls: Type[Enum], _):
        pass

    def literal(self, values: Sequence[Any], _):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType, _):
        self.visit(key_type)
        self.visit(value_type)

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
        _,
    ):
        for cls in types.values():
            self.visit(cls)

    def new_type(self, cls: AnyType, super_type: AnyType, _):
        self.visit(super_type)

    def primitive(self, cls: Type, _):
        pass

    def subprimitive(self, cls: Type, superclass: Type, _):
        self.visit(superclass)

    def tuple(self, types: Sequence[AnyType], _):
        for cls in types:
            self.visit(cls)

    def typed_dict(self, cls: Type, keys: Mapping[str, AnyType], total: bool, _):
        for cls in keys.values():
            self.visit(cls)

    def _union_arg(self, cls: AnyType, _):
        return None

    def _union_result(self, results: Sequence, _):
        pass

    def visit(self, cls: AnyType, _=None):
        if self._is_conversion(cls):
            return self.visit_not_builtin(cls, _)
        # Annotated is not always hashable
        if is_hashable(cls):
            if self._incr_ref(get_ref(cls), cls):
                return
        super().visit(cls, _)



__all__ = [
    "NotNull",
    "Skip",
    "Unsupported",
    "ValidationError",
    "alias",
    "check_types",
    "conversion",
    "deserialization",
    "deserialize",
    "deserializer",
    "fields",
    "json_schema",
    "metadata",
    "properties",
    "schema",
    "schema_ref",
    "serialization",
    "serialize",
    "serializer",
    "validation",
    "validator",
]


from . import (
    conversion,
    deserialization,
    fields,
    json_schema,
    metadata,
    serialization,
    validation,
)
from .aliases import alias
from .conversion import deserializer, serializer
from .deserialization import deserialize
from .json_schema.refs import schema_ref
from .json_schema.schema import schema
from .metadata import properties
from .serialization import serialize
from .type_checker import check_types
from .types import NotNull, Skip
from .validation import (
    ValidationError,
    validator,
)
from .visitor import Unsupported

# Handle standard library + internal types
serializer(ValidationError.format, ValidationError)
from . import std_types  # noqa: E402

del std_types  # clean namespace



from contextlib import suppress
from dataclasses import dataclass, fields
from typing import Any, Mapping, Optional, Sequence, TypeVar, overload

from apischema.types import AnyType, MetadataMixin, Number
from apischema.utils import Nil, as_dict, merge_opts, merge_opts_mapping, to_camel_case
from .annotations import (
    Annotations,
    _annotations,
    get_annotations,
)
from .constraints import (
    ArrayConstraints,
    Constraints,
    NumberConstraints,
    ObjectConstraints,
    StringConstraints,
    _constraints,
    get_constraints,
)

T = TypeVar("T")


@dataclass(frozen=True)
class Schema(MetadataMixin):
    annotations: Optional[Annotations] = None
    constraints: Optional[Constraints] = None

    def __post_init__(self):
        from apischema.metadata.keys import SCHEMA_METADATA

        super().__init__(SCHEMA_METADATA)

    def __call__(self, obj: T) -> T:
        if self.annotations is not None:
            _annotations[obj] = self.annotations
        if self.constraints is not None:
            _constraints[obj] = self.constraints
        return obj

    @property
    def override(self) -> bool:
        return self.annotations is not None and self.annotations.extra_only


_constraint_rewrite = {
    "min": "minimum",
    "max": "maximum",
    "exc_min": "exclusive_minimum",
    "exc_max": "exclusive_maximum",
    "mult_of": "multiple_of",
    "min_len": "min_length",
    "max_len": "max_length",
    "unique": "unique_items",
    "min_props": "min_properties",
    "max_props": "max_properties",
}


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Nil,
    examples: Optional[Sequence[Any]] = None,
    extra: Mapping[str, Any] = None,
    extra_only: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Nil,
    examples: Optional[Sequence[Any]] = None,
    min: Optional[Number] = None,
    max: Optional[Number] = None,
    exc_min: Optional[Number] = None,
    exc_max: Optional[Number] = None,
    mult_of: Optional[Number] = None,
    extra: Mapping[str, Any] = None,
    extra_only: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Nil,
    examples: Optional[Sequence[Any]] = None,
    format: Optional[str] = None,
    min_len: Optional[int] = None,
    max_len: Optional[int] = None,
    pattern: Optional[str] = None,
    extra: Mapping[str, Any] = None,
    extra_only: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Nil,
    examples: Optional[Sequence[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique: Optional[bool] = None,
    extra: Mapping[str, Any] = None,
    extra_only: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Nil,
    examples: Optional[Sequence[Any]] = None,
    min_props: Optional[int] = None,
    max_props: Optional[int] = None,
    extra: Mapping[str, Any] = None,
    extra_only: bool = False,
) -> Schema:
    ...


def schema(**kwargs) -> Schema:
    kwargs_ = {_constraint_rewrite.get(k, k): v for k, v in kwargs.items()}
    annotations_fields = {f.name for f in fields(Annotations)}
    annotations_kwargs = {k: v for k, v in kwargs_.items() if k in annotations_fields}
    constraint_kwargs = {
        k: v for k, v in kwargs_.items() if k not in annotations_kwargs
    }
    annotations = None
    if annotations_kwargs:
        annotations = Annotations(**annotations_kwargs)

    constraints = None
    if constraint_kwargs:
        for cls in (
            NumberConstraints,
            StringConstraints,
            ArrayConstraints,
            ObjectConstraints,
        ):
            with suppress(TypeError):
                constraints = cls(**constraint_kwargs)
                break
        else:
            raise TypeError("Invalid constraints")

    return Schema(annotations, constraints)


def get_schema(cls: AnyType) -> Schema:
    return Schema(get_annotations(cls), get_constraints(cls))


@merge_opts
def merge_annotations(default: Annotations, override: Annotations) -> Annotations:
    return Annotations(
        **{
            **as_dict(default),
            **as_dict(override),
            "extra": merge_opts_mapping(default.extra, override.extra),  # type: ignore
            "extra_only": default.extra_only or override.extra_only,
        }
    )


@merge_opts
def merge_constraints(default: Constraints, override: Constraints) -> Constraints:
    return default.merge(override)


@merge_opts
def merge_schema(default: Schema, override: Schema) -> Schema:
    return Schema(
        merge_annotations(default.annotations, override.annotations),
        merge_constraints(default.constraints, override.constraints),
    )


def _camel_case_and_remove_none(obj: Any) -> Mapping[str, Any]:
    if obj is None:
        return {}
    return {to_camel_case(k): v for k, v in as_dict(obj).items() if v is not None}


def serialize_schema(schema: Schema) -> Mapping[str, Any]:
    result = {
        **_camel_case_and_remove_none(schema.annotations),
        **_camel_case_and_remove_none(schema.constraints),
    }
    if schema.annotations is not None:
        if schema.annotations.default is Nil:
            result.pop("default")
        elif schema.annotations.default is None:
            result["default"] = None
    result.pop("extraOnly", ...)
    result.update(result.pop("extra", {}))
    return result



from dataclasses import MISSING, field, make_dataclass
from inspect import Parameter, signature
from typing import Any, Callable, List, Mapping, TypeVar

from apischema.conversion.converters import deserializer
from apischema.conversion.utils import Conversions, Converter
from apischema.fields import with_fields_set
from apischema.typing import get_type_hints
from apischema.utils import MakeDataclassField, as_dict, to_camel_case


def to_raw_deserializer(func: Callable) -> Converter:
    types = get_type_hints(func, include_extras=True)
    if "return" not in types:
        raise TypeError("Return must be annotated")
    sig = signature(func)
    fields: List[MakeDataclassField] = []
    kwargs = None
    for name, param in sig.parameters.items():
        if param.kind == Parameter.POSITIONAL_ONLY:  # pragma: no cover
            raise TypeError("Forbidden positional-only parameter")
        if param.kind == Parameter.VAR_POSITIONAL:
            raise TypeError("Forbidden variadic positional parameter")
        if param.kind == Parameter.VAR_KEYWORD:
            from apischema import properties

            field_ = field(default_factory=dict, metadata=properties)
            type_ = Mapping[str, types.get(name, Any)]  # type: ignore
            fields.append((name, type_, field_))  # type: ignore
            kwargs = name
            continue
        default = param.default if param.default is not Parameter.empty else MISSING
        try:
            fields.append((name, types[name], field(default=default)))
        except KeyError:
            raise TypeError("All parameters must be annotated")

    def converter(obj):
        kw = as_dict(obj)
        kw.update(kw.pop(kwargs, ()))
        return func(**kw)

    cls = with_fields_set(make_dataclass(to_camel_case(func.__name__), fields))
    converter.__annotations__ = {"obj": cls, "return": types["return"]}
    return converter


Func = TypeVar("Func", bound=Callable)


def raw_deserializer(func: Func, conversions: Conversions = None) -> Func:
    deserializer(to_raw_deserializer(func), conversions=conversions)
    return func



import re
from dataclasses import InitVar, dataclass, field, fields
from math import gcd
from operator import or_
from typing import (
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Pattern,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.types import Number
from apischema.utils import to_hashable
from apischema.validation.errors import ValidationError

MERGE_METADATA = "merge"

T = TypeVar("T")


def merge(operation: Callable[[Any, Any], Any], *, init=True) -> Optional[Any]:
    return field(default=None, init=init, metadata={MERGE_METADATA: operation})


_Constraints = TypeVar("_Constraints", bound="Constraints")


@dataclass(frozen=True)
class Constraints:
    valid_types: ClassVar[Union[Type, Tuple[Type, ...]]] = object
    _cache: ClassVar[Dict["Constraints", "Constraints"]]

    def _validate(self, data) -> Iterator[str]:
        yield from ()

    def validate(self, data: T) -> T:
        errors = list(self._validate(data))
        if errors:
            raise ValidationError(errors)
        return data

    def merge(self: _Constraints, other: Optional[_Constraints]) -> _Constraints:
        if other is None:
            return self
        try:
            return self._cache[other]  # type: ignore
        except (KeyError, AttributeError):
            if type(other) != type(self):
                raise TypeError("Incompatible constraints types")
            constraints: Dict[str, Any] = {}
            for field_ in fields(self):
                if getattr(self, field_.name) is None:
                    constraints[field_.name] = getattr(other, field_.name)
                elif getattr(other, field_.name) is None:
                    constraints[field_.name] = getattr(self, field_.name)
                else:
                    constraints[field_.name] = field_.metadata[MERGE_METADATA](
                        getattr(self, field_.name), getattr(other, field_.name)
                    )
            result = type(self)(**constraints)  # type: ignore
            if not hasattr(self, "_cache"):
                super().__setattr__("_cache", {})
            self._cache[other] = result
            return result


_constraints: Dict[Any, Constraints] = {}

get_constraints = _constraints.get

Cls = TypeVar("Cls", bound=type)


@dataclass(frozen=True)
class NumberConstraints(Constraints):
    valid_types = (int, float)
    minimum: Optional[Number] = merge(max)
    maximum: Optional[Number] = merge(min)
    exclusive_minimum: Optional[Number] = merge(min)
    exclusive_maximum: Optional[Number] = merge(max)
    multiple_of: Optional[Number] = merge(lambda m1, m2: m1 * m2 / gcd(m1, m2))

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, self.valid_types)
        if self.minimum is not None and data < self.minimum:
            yield f"less than {self.minimum} (minimum)"
        if self.maximum is not None and data > self.maximum:
            yield f"greater than {self.maximum} (maximum)"
        if self.exclusive_minimum is not None and data <= self.exclusive_minimum:
            yield f"less than or equal to {self.exclusive_minimum} (exclusiveMinimum)"
        if self.exclusive_maximum is not None and data >= self.exclusive_maximum:
            yield f"greater than or equal to {self.exclusive_maximum} (exclusiveMaximum)"  # noqa: E501
        if self.multiple_of is not None and (data % self.multiple_of) != 0:
            yield f"not a multiple of {self.multiple_of} (multipleOf)"


def merge_pattern(p1: Pattern, p2: Pattern) -> Pattern:
    raise TypeError("Cannot merge patterns")


@dataclass(frozen=True)
class StringConstraints(Constraints):
    valid_types = str
    min_length: Optional[int] = merge(max)
    max_length: Optional[int] = merge(min)
    pattern: InitVar[Optional[Union[str, Pattern]]] = merge(merge_pattern)
    _pattern: Optional[Pattern] = field(init=False)

    def __post_init__(self, pattern: Optional[Union[str, Pattern]]):
        super().__setattr__(
            "_pattern", re.compile(pattern) if pattern is not None else None
        )

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, self.valid_types)
        if self.min_length is not None and len(data) < self.min_length:
            yield f"length less than {self.min_length} (minLength)"
        if self.max_length is not None and len(data) > self.max_length:
            yield f"length greater than {self.max_length} (maxLength)"
        if self._pattern is not None and not re.match(self._pattern, data):
            yield f"unmatched pattern '{self._pattern.pattern}'"


@dataclass(frozen=True)
class ArrayConstraints(Constraints):
    valid_types = Collection
    min_items: Optional[int] = merge(max)
    max_items: Optional[int] = merge(min)
    unique_items: Optional[bool] = merge(or_)

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, self.valid_types)
        if self.min_items is not None and len(data) < self.min_items:
            yield f"size less than {self.min_items} (minItems)"
        if self.max_items is not None and len(data) > self.max_items:
            yield f"size greater than {self.max_items} (maxItems)"
        if self.unique_items is not None:
            if len(set(map(to_hashable, data))) != len(data):
                yield "duplicate items (uniqueItems)"


@dataclass(frozen=True)
class ObjectConstraints(Constraints):
    valid_types = Mapping
    min_properties: Optional[int] = merge(max)
    max_properties: Optional[int] = merge(min)

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, self.valid_types)
        if self.min_properties is not None and len(data) < self.min_properties:
            yield f"size less than {self.min_properties} (minProperties)"
        if self.max_properties is not None and len(data) > self.max_properties:
            yield f"size greater than {self.max_properties} (maxProperties)"



from contextlib import contextmanager
from dataclasses import is_dataclass
from enum import Enum
from itertools import chain
from typing import (
    Any,
    Callable,
    Collection,
    ContextManager,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from apischema.dataclasses.cache import get_aggregate_serialization_fields
from apischema.fields import fields_set
from apischema.json_schema.constraints import get_constraints
from apischema.types import AnyType
from apischema.typing import get_type_hints
from apischema.utils import type_name
from apischema.validation import get_validators, validate
from apischema.validation.errors import (
    ErrorKey,
    ErrorMsg,
    ValidationError,
    merge_errors,
)
from apischema.visitor import Visitor


def check_type(obj: Any, cls: AnyType):
    if not isinstance(obj, cls):
        raise ValidationError(
            [f"expected {type_name(cls)}, found type {type_name(type(obj))}"]
        )


@contextmanager
def catcher() -> Iterator[Callable[[ErrorKey], ContextManager]]:
    @contextmanager
    def catch(key: ErrorKey = None):
        try:
            yield
        except ValidationError as err:
            if key is not None:
                errors[key] = merge_errors(errors.get(key), err)
            else:
                msgs.extend(err.messages)

    msgs: List[ErrorMsg] = []
    errors: Dict[ErrorKey, ValidationError] = {}
    yield catch
    if errors:
        raise ValidationError(children=errors)


class TypeChecker(Visitor[Any, Any]):
    def __init__(self, validate: bool):
        super().__init__()
        self.validate = validate

    def any(self, obj):
        pass

    def collection(self, cls: Type[Collection], value_type: AnyType, obj):
        check_type(obj, cls)
        errors: Dict[ErrorKey, ValidationError] = {}
        for i, elt in enumerate(obj):
            try:
                self.visit(value_type, elt)
            except ValidationError as err:
                errors[i] = err
        if errors:
            raise ValidationError(children=errors)

    def dataclass(self, cls: Type, obj):
        assert is_dataclass(cls)
        check_type(obj, cls)
        types = get_type_hints(cls)
        errors: Dict[ErrorKey, ValidationError] = {}
        fields, aggregate_fields = get_aggregate_serialization_fields(cls)
        fields_set_ = fields_set(obj)
        for field in chain(fields, aggregate_fields):
            if field.name not in fields_set_:
                continue
            try:
                attr = getattr(obj, field.name)
                self.visit(types[field.name], attr)
                if self.validate:
                    if attr is not None and field.constraints is not None:
                        field.constraints.validate(attr)
                    if field.validators:
                        validate(attr, field.validators)
            except ValidationError as err:
                errors[field.name] = err
        if errors:
            raise ValidationError(children=errors)
        constraints = get_constraints(cls)
        if constraints is not None:
            constraints.validate(obj)
        validate(obj, [v for v in get_validators(cls) if not v.params])

    def enum(self, cls: Type[Enum], obj):
        check_type(obj, cls)

    def literal(self, values: Sequence[Any], obj):
        if obj not in values:
            raise ValidationError(f"value is not one of {values}")

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType, obj):
        check_type(obj, cls)
        errors: Dict[ErrorKey, ValidationError] = {}
        for k, v in obj.items():
            try:
                self.visit(key_type, k)
            except ValidationError as err:
                errors[k] = err
            try:
                self.visit(value_type, v)
            except ValidationError as err:
                errors[k] = merge_errors(errors.get(k), err)
        if errors:
            raise ValidationError(children=errors)

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
        obj,
    ):
        check_type(obj, cls)
        errors: Dict[ErrorKey, ValidationError] = {}
        for field, field_type in types.items():
            try:
                self.visit(field_type, getattr(obj, field))
            except ValidationError as err:
                errors[field] = err
        if errors:
            raise ValidationError(children=errors)
        constraints = get_constraints(cls)
        if constraints is not None:
            constraints.validate(obj)
        validate(obj)

    def primitive(self, cls: Type, obj):
        check_type(obj, cls)

    def subprimitive(self, cls: Type, superclass: Type, obj):
        check_type(obj, cls)
        constraints = get_constraints(cls)
        if constraints is not None:
            constraints.validate(obj)
        validate(obj)

    def tuple(self, types: Sequence[AnyType], obj):
        check_type(obj, tuple)
        if len(obj) != len(types):
            raise ValidationError(
                [f"expected tuple of length {len(types)}, found {len(obj)}"]
            )
        errors: Dict[ErrorKey, ValidationError] = {}
        for (i, type_), elt in zip(enumerate(types), obj):
            try:
                self.visit(type_, elt)
            except ValidationError as err:
                errors[i] = err
        if errors:
            raise ValidationError(children=errors)

    def typed_dict(self, cls: Type, keys: Mapping[str, AnyType], total: bool, obj):
        check_type(obj, Mapping)
        errors: Dict[ErrorKey, ValidationError] = {}
        for key, type_ in keys.items():
            if key in obj:
                try:
                    self.visit(type_, obj[key])
                except ValidationError as err:
                    errors[key] = err
            elif total:
                errors[key] = ValidationError("missing property")
        if errors:
            raise ValidationError(children=errors)
        constraints = get_constraints(cls)
        if constraints is not None:
            constraints.validate(obj)
        validate(obj, get_validators(cls))

    def union(self, alternatives: Sequence[AnyType], obj):
        error: Optional[ValidationError] = None
        for cls in alternatives:
            try:
                return self.visit(cls, obj)
            except ValidationError as err:
                error = merge_errors(error, err)
        raise error if error is not None else RuntimeError("Empty union")

    def unsupported(self, cls: AnyType, obj):
        check_type(obj, cls)


T = TypeVar("T")


def check_types(cls: AnyType, obj: T, *, validate: bool = False) -> T:
    TypeChecker(validate).visit(cls, obj)
    return obj



from typing import (  # type: ignore
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from apischema.conversion.converters import (
    _deserializers,
    _extra_deserializers,
    _extra_serializers,
    _serializers,
)
from apischema.conversion.utils import Conversions, ConverterWithConversions
from apischema.visitor import (
    Arg,
    Return,
    Visitor,
)

Conv = TypeVar("Conv")
Deserialization = Mapping[Type, ConverterWithConversions]
Serialization = Tuple[Type, ConverterWithConversions]


class ConversionsVisitor(Generic[Conv, Arg, Return], Visitor[Arg, Return]):
    def __init__(self, conversions: Optional[Conversions]):
        super().__init__()
        self.conversions = conversions

    @staticmethod
    def is_conversion(cls: Type, conversions: Optional[Conversions]) -> Optional[Conv]:
        raise NotImplementedError()

    def visit_conversion(self, cls: Type, conversion: Conv, arg: Arg) -> Return:
        raise NotImplementedError()

    # For typing
    def visit_not_conversion(self, cls: Type, arg: Arg) -> Return:
        ...

    visit_not_conversion = Visitor.visit_not_builtin  # noqa F811

    def visit_not_builtin(self, cls: Type, arg: Arg) -> Return:
        conversions = self.conversions
        conversion = self.is_conversion(cls, conversions)
        if conversion is None:
            self.conversions = None
            try:
                return self.visit_not_conversion(cls, arg)
            finally:
                self.conversions = conversions
        else:
            return self.visit_conversion(cls, conversion, arg)


class DeserializationVisitor(ConversionsVisitor[Deserialization, Arg, Return]):
    @staticmethod
    def is_conversion(
        cls: Type, conversions: Optional[Conversions]
    ) -> Optional[Deserialization]:
        if conversions is not None and cls in _extra_deserializers:
            try:  # cannot use __contains__ because of defaultdict/etc.
                sources = conversions[cls]
            except KeyError:
                pass
            else:
                if not isinstance(sources, Sequence):
                    sources = [sources]
                try:
                    return {
                        source: _extra_deserializers[cls][source] for source in sources
                    }
                except KeyError:
                    pass
        return _deserializers.get(cls)


class SerializationVisitor(ConversionsVisitor[Serialization, Arg, Return]):
    @staticmethod
    def is_conversion(
        cls: Type, conversions: Optional[Conversions]
    ) -> Optional[Serialization]:
        if not hasattr(cls, "__mro__"):
            return None
        for sub_cls in cls.__mro__:
            if conversions is not None and sub_cls in _extra_serializers:
                try:  # cannot use __contains__ because of defaultdict/etc.
                    target = conversions[sub_cls]
                except KeyError:
                    pass
                else:
                    if target is sub_cls:
                        return None
                    try:
                        return target, _extra_serializers[sub_cls][target]
                    except KeyError:
                        pass
            if sub_cls in _serializers:
                return _serializers[sub_cls]
        return None



from collections import defaultdict
from contextlib import contextmanager
from enum import Enum
from itertools import chain
from typing import (  # type: ignore
    AbstractSet,
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from apischema import settings
from apischema.conversion.utils import Conversions
from apischema.conversion.visitor import Conv
from apischema.dataclasses import is_dataclass, replace
from apischema.dataclasses.cache import Field, FieldKind
from apischema.json_schema.constraints import (
    ArrayConstraints,
    Constraints,
    NumberConstraints,
    ObjectConstraints,
    StringConstraints,
)
from apischema.json_schema.generation.refs import Refs, RefsExtractor
from apischema.json_schema.generation.visitor import (
    DeserializationSchemaVisitor,
    SchemaVisitor,
    SerializationSchemaVisitor,
)
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.json_schema.schema import (
    Schema,
    get_schema,
    merge_schema,
    serialize_schema,
)
from apischema.json_schema.types import JsonSchema, JsonType, json_schema
from apischema.json_schema.versions import JsonSchemaVersion, RefFactory
from apischema.metadata.keys import is_aggregate_field
from apischema.serialization import serialize
from apischema.types import AnyType
from apischema.utils import get_default, is_hashable

constraint_by_type = {
    int: NumberConstraints,
    float: NumberConstraints,
    str: StringConstraints,
}


def check_constraints(schema: Optional[Schema], expected: Type[Constraints]):
    if schema is not None and schema.constraints is not None:
        if not isinstance(schema.constraints, expected):
            raise TypeError(
                f"Bad constraints: expected {expected.__name__}"
                f" found {type(schema.constraints).__name__}"
            )


def check_no_constraints(schema: Optional[Schema], type_):
    if schema is not None and schema.constraints is not None:
        raise TypeError(f"{type} cannot have constraints")


def _with_schema(schema: Optional[Schema], json_schema: JsonSchema) -> JsonSchema:
    if schema is None:
        return json_schema
    elif schema.override:
        return JsonSchema(**serialize_schema(schema))
    else:
        return JsonSchema(json_schema, **serialize_schema(schema))


class SchemaBuilder(SchemaVisitor[Conv, Optional[Schema], JsonSchema]):
    def __init__(
        self,
        conversions: Optional[Conversions],
        ref_factory: RefFactory,
        refs: Collection[str],
        ignore_first_ref: bool,
    ):
        super().__init__(conversions)
        self.ref_factory = ref_factory
        self.refs = refs
        self.ignore_first_ref = ignore_first_ref

    def _ref_schema(self, ref: str, schema: Optional[Schema]) -> JsonSchema:
        return _with_schema(schema, JsonSchema({"$ref": self.ref_factory(ref)}))

    def _annotated(
        self, cls: AnyType, annotations: Sequence[Any], schema: Optional[Schema]
    ) -> JsonSchema:
        for annotation in reversed(annotations):
            if isinstance(annotation, schema_ref):
                annotation.check_type(cls)
                if annotation.ref in self.refs:
                    if self.ignore_first_ref:
                        self.ignore_first_ref = False
                    else:
                        assert isinstance(annotation.ref, str)
                        return self._ref_schema(annotation.ref, schema)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
            if isinstance(annotation, Schema):
                schema = merge_schema(annotation, schema)
        return self.visit(cls, schema)

    def any(self, schema: Optional[Schema]) -> JsonSchema:
        check_no_constraints(schema, "Any")
        return _with_schema(schema, JsonSchema())

    def collection(
        self, cls: Type[Iterable], value_type: AnyType, schema: Optional[Schema]
    ) -> JsonSchema:
        check_constraints(schema, ArrayConstraints)
        return _with_schema(
            schema,
            json_schema(
                type=JsonType.ARRAY,
                items=self.visit(value_type),
                uniqueItems=issubclass(cls, AbstractSet),
            ),
        )

    @staticmethod
    def _override_arg(cls: AnyType, schema: Optional[Schema]) -> Optional[Schema]:
        return merge_schema(get_schema(cls), schema)

    def _field_schema(self, field: Field) -> Schema:
        annotations = field.annotations
        if annotations is not None and annotations.default is ...:
            if not field.default:
                raise TypeError("Invalid ... without field default")
            try:
                default = serialize(get_default(field.base_field))
            except Exception:
                pass
            else:
                annotations = replace(annotations, default=default)
        return Schema(annotations, field.constraints)

    def _visit_field(self, field: Field) -> JsonSchema:
        result = self._field_visit(field, self._field_schema(field))
        if not is_aggregate_field(field.base_field):
            result = json_schema(
                readOnly=field.kind == FieldKind.NO_INIT,
                writeOnly=field.kind == FieldKind.INIT,
                **result,
            )
        return result

    def _properties_schema(self, field: Field) -> JsonSchema:
        props_schema = JsonSchema()
        with self._without_ref():
            props_schema = self._visit_field(field)
        if not props_schema.get("type") == JsonType.OBJECT:
            raise TypeError("properties field must have an 'object' type")
        if "patternProperties" in props_schema:
            if (
                len(props_schema["patternProperties"]) != 1
                or "additionalProperties" in props_schema
            ):  # don't try to merge the schemas
                pass
            else:
                return next(iter(props_schema["patternProperties"].values()))
        elif "additionalProperties" in props_schema:
            if isinstance(props_schema["additionalProperties"], JsonSchema):
                return props_schema["additionalProperties"]
            else:  # there is maybe only properties
                pass
        return JsonSchema()

    _required: bool

    def dataclass(self, cls: Type, schema: Optional[Schema]) -> JsonSchema:
        assert is_dataclass(cls)
        check_constraints(schema, ObjectConstraints)
        properties = {}
        required: List[str] = []
        (
            fields,
            merged_fields,
            pattern_fields,
            additional_field,
        ) = self._dataclass_fields(cls)
        dependent: Dict[str, Set[str]] = defaultdict(set)
        for field in fields:
            properties[field.alias] = self._visit_field(field)
            if not field.default:
                required.append(field.alias)
            for req in self._required_by(field):
                dependent[req].add(field.alias)
        dependent_required = {
            field: sorted(dependent[field]) for field in sorted(dependent)
        }

        merged_schemas = [self._visit_field(field) for _, field in merged_fields]
        pattern_properties = {
            cast(Pattern, pattern)
            if pattern is not ...
            else infer_pattern(field): self._properties_schema(field)
            for pattern, field in pattern_fields
        }
        additional_properties: Union[bool, JsonSchema]
        if additional_field:
            additional_properties = self._properties_schema(additional_field)
        else:
            additional_properties = settings.additional_properties
        result = json_schema(
            type=JsonType.OBJECT,
            properties=properties,
            required=required,
            additionalProperties=additional_properties,
            patternProperties=pattern_properties,
            dependentRequired=dependent_required,
        )
        if merged_schemas:
            result = json_schema(
                type=JsonType.OBJECT,
                allOf=[result, *merged_schemas],
                unevaluatedProperties=False,
            )
        return _with_schema(schema, result)

    def enum(self, cls: Type[Enum], schema: Optional[Schema]) -> JsonSchema:
        check_no_constraints(schema, "Enum")
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls), schema)

    def literal(self, values: Sequence[Any], schema: Optional[Schema]) -> JsonSchema:
        check_no_constraints(schema, "Literal")
        if not values:
            raise TypeError("Empty Literal")
        types = sorted(
            set(
                JsonType.from_type(type(v.value if isinstance(v, Enum) else v))
                for v in values
            )
        )
        # Mypy issue
        type_: Any = types[0] if len(types) == 1 else types
        return _with_schema(
            schema,
            json_schema(type_=type_, enum=values)
            if len(values) != 1
            else json_schema(type=type_, const=values[0]),
        )

    def mapping(
        self,
        cls: Type[Mapping],
        key_type: AnyType,
        value_type: AnyType,
        schema: Optional[Schema],
    ) -> JsonSchema:
        check_constraints(schema, ObjectConstraints)
        with self._without_ref():
            key = self.visit(key_type)
        if key["type"] != JsonType.STRING:
            raise ValueError("Mapping types must string-convertible key")
        value = self.visit(value_type)
        return _with_schema(
            schema,
            json_schema(type=JsonType.OBJECT, patternProperties={key["pattern"]: value})
            if "pattern" in key
            else json_schema(type=JsonType.OBJECT, additionalProperties=value),
        )

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
        schema: Optional[Schema],
    ) -> JsonSchema:
        check_constraints(schema, ObjectConstraints)
        return _with_schema(
            schema,
            json_schema(
                type=JsonType.OBJECT,
                properties={key: self.visit(key) for key, cls in types.items()},
                required=sorted(types.keys() - defaults.keys()),
                additionalProperties=settings.additional_properties,
            ),
        )

    def new_type(
        self, cls: AnyType, super_type: AnyType, schema: Optional[Schema]
    ) -> JsonSchema:
        return self.visit(super_type, schema)

    def primitive(self, cls: Type, schema: Optional[Schema]) -> JsonSchema:
        if cls in constraint_by_type:
            check_constraints(schema, constraint_by_type[cls])
        return _with_schema(schema, JsonSchema(type=JsonType.from_type(cls)))

    def subprimitive(
        self, cls: Type, superclass: Type, schema: Optional[Schema]
    ) -> JsonSchema:
        return self.primitive(superclass, schema)

    def tuple(self, types: Sequence[AnyType], schema: Optional[Schema]) -> JsonSchema:
        check_constraints(schema, ArrayConstraints)
        if schema is not None and schema.constraints is not None:
            assert isinstance(schema.constraints, ArrayConstraints)
            if (
                schema.constraints.max_items is not None
                or schema.constraints.min_items is not None
            ):
                raise TypeError("Tuple cannot have min_items/max_items constraints")
        return _with_schema(
            schema,
            json_schema(
                type=JsonType.ARRAY,
                items=[self.visit(cls) for cls in types],
                minItems=len(types),
                maxItems=len(types),
            ),
        )

    def typed_dict(
        self,
        cls: Type,
        keys: Mapping[str, AnyType],
        total: bool,
        schema: Optional[Schema],
    ) -> JsonSchema:
        check_constraints(schema, ObjectConstraints)
        return _with_schema(
            schema,
            json_schema(
                type=JsonType.OBJECT,
                properties={key: self.visit(cls) for key, cls in keys.items()},
                required=list(keys) if total else [],
            ),
        )

    def _union_arg(self, cls: AnyType, arg: Optional[Schema]) -> Optional[Schema]:
        return None

    def _union_result(
        self, results: Sequence[JsonSchema], schema: Optional[Schema]
    ) -> JsonSchema:
        if len(results) == 1:
            result = results[0]
        elif all(alt.keys() == {"type"} for alt in results):
            types = chain.from_iterable(
                [res["type"]] if isinstance(res["type"], JsonType) else res["type"]
                for res in results
            )
            result = json_schema(type=list(types))
        else:
            result = json_schema(anyOf=results)
        return _with_schema(schema, result)

    def visit(self, cls: AnyType, schema: Optional[Schema] = None) -> JsonSchema:
        if self._is_conversion(cls):
            return self.visit_not_builtin(cls, merge_schema(get_schema(cls), schema))
        if is_hashable(cls):
            ref = get_ref(cls)
            if ref in self.refs:
                if self.ignore_first_ref:
                    self.ignore_first_ref = False
                else:
                    assert isinstance(ref, str)
                    return self._ref_schema(ref, schema)
            schema = merge_schema(get_schema(cls), schema)
        return super().visit(cls, schema)

    @contextmanager
    def _without_ref(self):
        refs_save = self.refs
        self.refs = ()
        try:
            yield
        finally:
            self.refs = refs_save

    RefsExtractor: Type["RefsExtractor"]


class DeserializationSchemaBuilder(DeserializationSchemaVisitor, SchemaBuilder):
    class RefsExtractor(DeserializationSchemaVisitor, RefsExtractor):  # type: ignore
        pass


class SerializationSchemaBuilder(SerializationSchemaVisitor, SchemaBuilder):
    class RefsExtractor(SerializationSchemaVisitor, RefsExtractor):  # type: ignore
        pass


TypesWithConversions = Sequence[Union[AnyType, Tuple[AnyType, Conversions]]]


def _default_version(
    version: Optional[JsonSchemaVersion],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
) -> Tuple[JsonSchemaVersion, RefFactory, bool]:
    if version is None:
        version = settings.json_schema_version
    if ref_factory is None:
        ref_factory = version.ref_factory
    if all_refs is None:
        all_refs = version.all_refs
    return version, ref_factory, all_refs


def _export_refs(
    types: TypesWithConversions, builder: Type[SchemaBuilder], all_refs: bool
) -> Mapping[str, AnyType]:
    refs: Refs = {}
    for cls in types:
        conversions = None
        if isinstance(cls, tuple):
            cls, conversions = cls
        builder.RefsExtractor(conversions, refs).visit(cls)
    filtr = (lambda count: True) if all_refs else (lambda count: count > 1)
    return {ref: cls for ref, (cls, count) in refs.items() if filtr(count)}


def _refs_schema(
    builder: Type[SchemaBuilder], refs: Mapping[str, AnyType], ref_factory: RefFactory
) -> Mapping[str, JsonSchema]:
    return {
        ref: builder(None, ref_factory, refs, True).visit(cls)
        for ref, cls in refs.items()
    }


def _schema(
    builder: Type[SchemaBuilder],
    cls: AnyType,
    schema: Optional[Schema],
    conversions: Optional[Conversions],
    version: Optional[JsonSchemaVersion],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
    with_schema: bool,
) -> Mapping[str, Any]:
    add_defs = ref_factory is None
    if ref_factory is not None and all_refs is None:
        all_refs = True
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    refs = _export_refs([(cls, conversions)], builder, all_refs)
    json_schema = builder(conversions, ref_factory, refs, False).visit(cls, schema)
    if add_defs:
        defs = _refs_schema(builder, refs, ref_factory)
        if defs:
            json_schema["$defs"] = defs
    result = serialize(json_schema, conversions=version.conversions)
    if with_schema and version.schema is not None:
        result["$schema"] = version.schema
    return result


def deserialization_schema(
    cls: AnyType,
    *,
    schema: Schema = None,
    conversions: Conversions = None,
    version: JsonSchemaVersion = None,
    ref_factory: RefFactory = None,
    all_refs: bool = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    return _schema(
        DeserializationSchemaBuilder,
        cls,
        schema,
        conversions,
        version,
        ref_factory,
        all_refs,
        with_schema,
    )


def serialization_schema(
    cls: AnyType,
    *,
    schema: Schema = None,
    conversions: Conversions = None,
    version: JsonSchemaVersion = None,
    ref_factory: RefFactory = None,
    all_refs: bool = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    return _schema(
        SerializationSchemaBuilder,
        cls,
        schema,
        conversions,
        version,
        ref_factory,
        all_refs,
        with_schema,
    )


def _defs_schema(
    types: TypesWithConversions,
    builder: Type[SchemaBuilder],
    ref_factory: RefFactory,
    all_refs: bool,
) -> Mapping[str, JsonSchema]:
    return _refs_schema(builder, _export_refs(types, builder, all_refs), ref_factory)


def _set_missing_properties(
    schema: JsonSchema, properties: Optional[Mapping[str, JsonSchema]], key: str,
) -> JsonSchema:
    if properties is None:
        return schema
    missing = {name: prop for name, prop in properties.items() if prop.get(key, False)}
    schema.setdefault("properties", {}).update(missing)
    return schema


def definitions_schema(
    *,
    deserialization: TypesWithConversions = (),
    serialization: TypesWithConversions = (),
    version: JsonSchemaVersion = None,
    ref_factory: Optional[RefFactory] = None,
    all_refs: bool = None,
) -> Mapping[str, Mapping[str, Any]]:
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    deserialization_schemas = _defs_schema(
        deserialization, DeserializationSchemaBuilder, ref_factory, all_refs
    )
    serialization_schemas = _defs_schema(
        serialization, SerializationSchemaBuilder, ref_factory, all_refs
    )
    for duplicate in deserialization_schemas.keys() & serialization_schemas.keys():
        d_schema = deserialization_schemas[duplicate]
        s_schema = serialization_schemas[duplicate]
        _set_missing_properties(
            s_schema, d_schema.get("properties"), "writeOnly",
        )
        _set_missing_properties(
            d_schema, s_schema.get("properties"), "readOnly",
        )
        if "required" in d_schema and "required" in s_schema:
            s_schema["required"] = d_schema["required"]
        if d_schema != s_schema:
            raise TypeError(
                f"Reference {duplicate} has different schemas"
                f" for deserialization and serialization"
            )
    return {
        ref: serialize(schema, conversions=version.conversions)
        for ref, schema in chain(
            deserialization_schemas.items(), serialization_schemas.items()
        )
    }



from dataclasses import dataclass
from typing import Any

from apischema.utils import (
    Nil,
    as_dict,
    to_camel_case,
    to_hashable,
)


def test_to_hashable():
    hashable1 = to_hashable({"key1": 0, "key2": [1, 2]})
    hashable2 = to_hashable({"key2": [1, 2], "key1": 0})
    assert hashable1 == hashable2
    assert hash(hashable1) == hash(hashable2)


def test_to_camel_case():
    assert to_camel_case("min_length") == "minLength"


@dataclass
class Data:
    default: Any = Nil


def test_as_dict():
    assert as_dict(Data()) == {"default": Nil}



import dataclasses
from enum import Enum, auto
from inspect import getmembers
from typing import (  # type: ignore
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    TYPE_CHECKING,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.conversion.conversions import ConversionsMetadata
from apischema.conversion.utils import (
    Conversions,
    check_converter,
    substitute_type_vars,
)
from apischema.conversion.visitor import (
    ConversionsVisitor,
    Deserialization,
    Serialization,
)
from apischema.dependencies import DependentRequired
from apischema.metadata.keys import (
    ALIAS_METADATA,
    CONVERSIONS_METADATA,
    DEFAULT_FALLBACK_METADATA,
    INCOMPATIBLE_WITH_MERGED,
    INCOMPATIBLE_WITH_PROPERTIES,
    INIT_VAR_METADATA,
    MERGED_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SCHEMA_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
)
from apischema.types import AnyType
from apischema.typing import get_type_hints
from apischema.utils import PREFIX
from apischema.validation.validator import VALIDATORS_ATTR, Validator, validate
from apischema.visitor import visitor_method

if TYPE_CHECKING:
    from apischema.json_schema.annotations import Annotations
    from apischema.json_schema.constraints import Constraints

Cls = TypeVar("Cls", bound=Type)


def slotted_dataclass(cls: Cls) -> Cls:
    slots = [f.name for f in dataclasses.fields(cls)]
    namespace = cls.__dict__.copy()
    for slot in slots:
        namespace.pop(slot, ...)
    namespace["__slots__"] = slots
    return cast(Cls, type(cls.__name__, cls.__bases__, namespace))


class FieldKind(Enum):
    INIT = auto()
    NORMAL = auto()
    NO_INIT = auto()

    def __and__(self, other: "FieldKind"):
        return (self != FieldKind.INIT and other != FieldKind.INIT) or (
            self != FieldKind.NO_INIT and other != FieldKind.NO_INIT
        )


@slotted_dataclass
@dataclasses.dataclass
class Field:
    alias: str
    annotations: Optional["Annotations"]
    base_field: dataclasses.Field
    constraints: Optional["Constraints"]
    default_fallback: bool
    default: bool
    deserialization_conversions: Optional[Conversions]
    deserialization_method: Callable
    deserialization_type: AnyType
    kind: FieldKind
    name: str
    post_init: bool
    serialization_conversions: Optional[Conversions]
    serialization_method: Callable
    serialization_type: AnyType
    type: AnyType
    validators: Optional[Sequence[Validator]]

    deserialization_required_by: AbstractSet[str] = frozenset()
    serialization_required_by: AbstractSet[str] = frozenset()


FieldCache = Tuple[
    Sequence[Field],
    Sequence[Tuple[AbstractSet[str], Field]],
    Sequence[Tuple[Union[Pattern, "ellipsis"], Field]],  # noqa F821
    Optional[Field],
]
_deserialization_fields: Dict[Type, FieldCache] = {}
AggregateFieldCache = Tuple[Sequence[Field], Sequence[Field]]
_aggregate_serialization_fields: Dict[Type, AggregateFieldCache] = {}
# Because dataclasses with InitVar doesn't have to have a __post_init__
# (they could only be used in validators), theses classes has to be flagged
_post_init_classes: Set[Type] = set()


def _to_aggregate(cache: FieldCache) -> AggregateFieldCache:
    fields, merged, pattern, additional = cache
    additional_ = [additional] if additional is not None else []
    return fields, [f for _, f in merged] + [f for _, f in pattern] + additional_


def _from_aggregate(aggregate_cache: AggregateFieldCache) -> FieldCache:
    fields, aggregate_fields = aggregate_cache
    merged_fields: List[Tuple[AbstractSet[str], Field]] = []
    pattern_fields: List[Tuple[Pattern, Field]] = []
    additional_fields = None
    for field in aggregate_fields:
        metadata = field.base_field.metadata
        if MERGED_METADATA in metadata:
            merged_fields.append((frozenset(_merged_aliases(field.type)), field))
        else:
            pattern = metadata[PROPERTIES_METADATA]
            if pattern is not None:
                pattern_fields.append((pattern, field))
            else:
                additional_fields = field
    return fields, merged_fields, pattern_fields, additional_fields


T = TypeVar("T")


def _add_field_to_lists(
    obj: T, kind: FieldKind, deserializations: List[T], serializations: List[T]
):
    if kind != FieldKind.INIT:
        serializations.append(obj)
    if kind != FieldKind.NO_INIT:
        deserializations.append(obj)


def _resolve_init_var(cls: AnyType, field: dataclasses.Field) -> AnyType:
    if INIT_VAR_METADATA not in field.metadata:
        raise TypeError("Before 3.8, InitVar requires init_var metadata")

    tmp_cls = dataclasses.make_dataclass(
        "Tmp",
        [
            (
                PREFIX,
                field.metadata[INIT_VAR_METADATA],
                dataclasses.field(default=cast(Any, None)),
            )
        ],
        bases=(cls,),
    )
    return get_type_hints(tmp_cls, include_extras=True)[PREFIX]


def _handle_method_conversions(
    method: Callable, conversions: Optional[Conversions]
) -> Callable:
    if conversions is None:
        return method

    def wrapper(self: ConversionsVisitor, *arg):
        conversions_save = self.conversions
        self.conversions = conversions
        try:
            return method(self, *arg)
        finally:
            self.conversions = conversions_save

    return wrapper


def _deserialization_method(
    deserialization_type: Type,
    deserialization: Optional[Deserialization],
    conversions: Optional[Conversions],
) -> Callable:
    from apischema.deserialization.deserializer import (
        Deserializer,
        DataWithConstraint,
    )

    if deserialization is None:
        return _handle_method_conversions(
            visitor_method(deserialization_type, Deserializer), conversions
        )
    else:

        def deserialization_method(  # type: ignore
            visitor: Deserializer, cls: AnyType, data2: DataWithConstraint
        ):  # type: ignore
            assert deserialization is not None
            return visitor.visit_conversion(cls, deserialization, data2)

        return deserialization_method


def _deserialization(
    field_type: AnyType, metadata: Mapping[str, Any],
) -> Tuple[
    AnyType,
    Optional[Deserialization],
    Optional[Conversions],
    Callable,
    Optional[Sequence[Validator]],
]:
    validators: Optional[Sequence[Validator]] = None
    if VALIDATORS_METADATA in metadata:
        validators = metadata[VALIDATORS_METADATA].validators
    conversions = metadata.get(CONVERSIONS_METADATA, ConversionsMetadata())
    deserialization_conversions: Optional[Conversions]
    deserialization: Optional[Deserialization]
    if conversions.both is not None:
        deserialization_conversions = {field_type: conversions.both}
    elif conversions.deserialization is not None:
        deserialization_conversions = conversions.deserialization
    else:
        deserialization_conversions = None
    # Embed validators in conversion in order to have only one if in deserialization
    # `if field.deserializer is not None`
    if conversions.deserializer is not None:
        converter = conversions.deserializer
        param, ret = check_converter(converter, None, field_type)  # type: ignore
        ret, param = substitute_type_vars(ret, param)  # type: ignore
        if validators:
            converter = lambda data, conv=converter: validate(  # noqa E731
                conv(data), validators
            )
        deserialization_type = param
        deserialization = {param: (converter, deserialization_conversions)}
    elif validators:
        deserialization_type = field_type
        deserialization = {
            field_type: ((lambda data: validate(data, validators)), None)
        }
    else:
        deserialization_type = field_type
        deserialization = None
    return (
        deserialization_type,
        deserialization,
        deserialization_conversions,
        _deserialization_method(
            field_type, deserialization, deserialization_conversions
        ),
        validators,
    )


def _serialization_method(
    serialization_type: Type,
    serialization: Optional[Serialization],
    conversions: Optional[Conversions],
) -> Callable:
    from apischema.serialization import (
        Serializer,
        PRIMITIVE_TYPES_SET,
        COLLECTION_TYPE_SET,
    )

    if serialization is None:
        # TODO could be optimized
        if serialization_type in PRIMITIVE_TYPES_SET:

            def serialization_method(visitor: Serializer, obj):
                return obj

        elif serialization_type in COLLECTION_TYPE_SET:

            def serialization_method(visitor: Serializer, obj):
                return [visitor.visit2(elt) for elt in obj]

        elif serialization_type is dict:

            def serialization_method(visitor: Serializer, obj):
                return {
                    visitor.visit2(key): visitor.visit2(value)
                    for key, value in obj.items()
                }

        else:
            serialization_method = Serializer.visit2
        return _handle_method_conversions(serialization_method, conversions)
    else:

        def serialization_method(visitor: Serializer, obj):  # type: ignore
            assert serialization is not None
            return visitor.visit_conversion(..., serialization, obj)

        return serialization_method


def _serialization(
    field_type: AnyType, metadata: Mapping[str, Any]
) -> Tuple[AnyType, Optional[Serialization], Optional[Conversions], Callable]:
    conversions = metadata.get(CONVERSIONS_METADATA, ConversionsMetadata())
    serialization_conversions: Optional[Conversions]
    serialization: Optional[Serialization]
    if conversions.both is not None:
        serialization_conversions = {field_type: conversions.both}
    elif conversions.deserialization is not None:
        serialization_conversions = conversions.deserialization
    else:
        serialization_conversions = None
    if conversions.serializer is not None:
        converter = conversions.serializer
        param, ret = check_converter(converter, field_type, None)  # type: ignore
        param, ret = substitute_type_vars(param, ret)  # type: ignore
        serialization_type = ret
        serialization = ret, (converter, serialization_conversions)
    else:
        serialization_type = field_type
        serialization = None
    return (
        serialization_type,
        serialization,
        serialization_conversions,
        _serialization_method(
            serialization_type, serialization, serialization_conversions
        ),
    )


def _merged_aliases(cls: Type) -> Iterable[str]:
    assert dataclasses.is_dataclass(cls)
    types = get_type_hints(cls, include_extras=True)
    for field in dataclasses.fields(cls):
        if MERGED_METADATA in field.metadata:
            # No need to check overlapping here because it will be checked
            # when merged dataclass will be cached
            yield from _merged_aliases(types[field.name])
        elif PROPERTIES_METADATA in field.metadata:
            raise TypeError("Merged dataclass cannot have properties field")
        else:
            yield field.metadata.get(ALIAS_METADATA, field.name)


def _check_fields_overlap(present: Set[str], other: AbstractSet[str]):
    if present & other:
        raise TypeError(f"Merged fields {present & other} overlap")
    present.update(other)


def _update_dependencies(cls: AnyType, all_fields: Mapping[str, Field]):
    for validator in getattr(cls, VALIDATORS_ATTR, ()):
        validator.dependencies = {
            dep for dep in validator.dependencies if dep in all_fields
        }
    all_dependencies: Collection[DependentRequired] = [
        m for _, m in getmembers(cls, lambda m: isinstance(m, DependentRequired))
    ]
    for dependencies in all_dependencies:
        for base_field, required in dependencies.required_by().items():
            field = all_fields[base_field.name]
            field.deserialization_required_by = {
                all_fields[req.name].alias
                for req in required
                if all_fields[req.name].kind != FieldKind.NO_INIT
            }
            field.serialization_required_by = {
                all_fields[req.name].alias
                for req in required
                if all_fields[req.name].kind != FieldKind.INIT
            }


F = TypeVar("F", bound=Union[Field, Tuple[Any, Field]])


def _filter_by_kind(field_list: Iterable[F], kind: FieldKind) -> Sequence[F]:
    fields = [(elt, elt[1] if isinstance(elt, tuple) else elt) for elt in field_list]
    return [elt for elt, field in fields if field.kind != kind]


@dataclasses.dataclass
class FieldLists:
    cls: Type
    normal: List[Field] = dataclasses.field(default_factory=list)
    merged: List[Tuple[AbstractSet, Field]] = dataclasses.field(default_factory=list)
    pattern: List[Tuple[Pattern, Field]] = dataclasses.field(default_factory=list)
    additional: List[Field] = dataclasses.field(default_factory=list)

    def remove_kind(self, remove: FieldKind) -> FieldCache:
        additional = _filter_by_kind(self.additional, remove)
        if len(additional) > 1:
            raise TypeError(
                f"{self.cls.__name__} cannot have more than one properties field"
            )
        return (
            _filter_by_kind(self.normal, remove),
            _filter_by_kind(self.merged, remove),
            _filter_by_kind(self.pattern, remove),
            additional[0] if additional else None,
        )


def cache_fields(cls: Type):
    assert dataclasses.is_dataclass(cls)
    types = get_type_hints(cls, include_extras=True)
    lists = FieldLists(cls)
    all_fields: Dict[str, Field] = {}
    all_merged_aliases: Set[str] = set()
    for field in getattr(cls, dataclasses._FIELDS).values():  # type: ignore
        if field._field_type == dataclasses._FIELD_CLASSVAR:  # type: ignore
            continue
        metadata = field.metadata
        if SKIP_METADATA in metadata:
            continue
        error_prefix = f"{cls.__name__}.{field.name}: "
        type_ = types[field.name]
        if isinstance(type_, dataclasses.InitVar):
            kind = FieldKind.INIT
            type_ = type_.type  # type: ignore
        elif type_ is dataclasses.InitVar:
            kind = FieldKind.INIT
            type_ = _resolve_init_var(cls, field)
        elif field.init:
            kind = FieldKind.NORMAL
        else:
            kind = FieldKind.NO_INIT
        if kind == FieldKind.INIT:
            _post_init_classes.add(cls)
        default = REQUIRED_METADATA not in metadata and (
            field.default is not dataclasses.MISSING
            or field.default_factory is not dataclasses.MISSING  # type: ignore
        )
        (
            deserialization_type,
            deserialization,
            deserialization_conversions,
            deserialization_method,
            validators,
        ) = _deserialization(type_, metadata)
        (
            serialization_type,
            serialization,
            serialization_conversions,
            serialization_method,
        ) = _serialization(type_, metadata)

        from apischema import settings
        from apischema.json_schema.schema import Schema

        schema = metadata.get(SCHEMA_METADATA, Schema())

        new_field = Field(
            alias=settings.aliaser()(metadata.get(ALIAS_METADATA, field.name)),
            annotations=schema.annotations,
            base_field=field,
            constraints=schema.constraints,
            default=default,
            default_fallback=metadata.get(DEFAULT_FALLBACK_METADATA, False),
            deserialization_conversions=deserialization_conversions,
            deserialization_method=deserialization_method,
            deserialization_type=deserialization_type,
            kind=kind,
            name=field.name,
            post_init=metadata.get(POST_INIT_METADATA, False),
            serialization_conversions=serialization_conversions,
            serialization_method=serialization_method,
            serialization_type=serialization_type,
            type=field.type,
            validators=validators,
        )
        all_fields[field.name] = new_field
        if MERGED_METADATA in metadata:
            if any(key in metadata for key in INCOMPATIBLE_WITH_MERGED):
                raise TypeError(f"{error_prefix}Incompatible metadata with merged")
            if not dataclasses.is_dataclass(type_):
                raise TypeError(
                    f"{error_prefix}Merged field must have a dataclass type"
                )
            merged_aliases: AbstractSet[str] = frozenset(_merged_aliases(type_))
            _check_fields_overlap(all_merged_aliases, merged_aliases)
            lists.merged.append((merged_aliases, new_field))
        elif PROPERTIES_METADATA in metadata:
            if any(key in metadata for key in INCOMPATIBLE_WITH_PROPERTIES):
                raise TypeError(f"{error_prefix}Incompatible metadata with properties")
            pattern = metadata[PROPERTIES_METADATA]
            if pattern is None:
                lists.additional.append(new_field)
            else:
                lists.pattern.append((pattern, new_field))
        else:
            lists.normal.append(new_field)
    _check_fields_overlap(all_merged_aliases, all_fields.keys())
    _update_dependencies(cls, all_fields)
    _deserialization_fields[cls] = lists.remove_kind(FieldKind.NO_INIT)
    _aggregate_serialization_fields[cls] = _to_aggregate(
        lists.remove_kind(FieldKind.INIT)
    )


def get_deserialization_fields(cls: Type) -> FieldCache:
    try:
        return _deserialization_fields[cls]
    except KeyError:
        cache_fields(cls)
        # Use recursion because of potential concurrent reset_dataclasses_cache
        return get_deserialization_fields(cls)


def get_serialization_fields(cls: Type) -> FieldCache:
    return _from_aggregate(get_aggregate_serialization_fields(cls))


def get_aggregate_serialization_fields(cls: Type) -> AggregateFieldCache:
    try:
        return _aggregate_serialization_fields[cls]
    except KeyError:
        cache_fields(cls)
        return get_aggregate_serialization_fields(cls)


def has_post_init_fields(cls: Type) -> bool:
    if cls not in _deserialization_fields:
        cache_fields(cls)
    return cls in _post_init_classes


def reset_dataclasses_cache():
    _deserialization_fields.clear()
    _aggregate_serialization_fields.clear()



import sys
from dataclasses import is_dataclass, replace as replace_

if sys.version_info <= (3, 7):
    is_dataclass_ = is_dataclass

    def is_dataclass(obj) -> bool:
        return is_dataclass_(obj) and getattr(obj, "__origin__", None) is None


def replace(*args, **changes):
    from apischema.fields import FIELDS_SET_ATTR, fields_set, set_fields

    result = replace_(*args, **changes)
    obj = args[0]
    if hasattr(obj, FIELDS_SET_ATTR):
        set_fields(result, *fields_set(obj), *changes, overwrite=True)
    return result



"""Kind of typing_extensions for this package"""
__all__ = [
    "Annotated",
    "Literal",
    "TypedDict",
    "_AnnotatedAlias",
    "_GenericAlias",
    "_LiteralMeta",
    "_TypedDictMeta",
    "get_type_hints",
]

import sys
from types import ModuleType
from typing import Any, Generic, TypeVar


class NoType:
    def __getitem__(self, key):  # pragma: no cover
        return self


NO_TYPE: Any = NoType()

if sys.version_info >= (3, 9):  # pragma: no cover
    from typing import Annotated, get_type_hints
else:  # pragma: no cover
    try:
        from typing_extensions import Annotated
    except ImportError:
        Annotated = NO_TYPE
    try:
        from typing_extensions import get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as _gth

        def gth(  # type: ignore
            obj, globalns=None, localns=None, include_extras=False
        ):
            return _gth(obj, globalns, localns)

    def get_type_hints(  # type: ignore
        obj, globalns=None, localns=None, include_extras=False
    ):
        # fix https://bugs.python.org/issue37838
        if not isinstance(obj, (type, ModuleType)) and globalns is None:
            nsobj = obj
            while hasattr(nsobj, "__wrapped__"):
                nsobj = nsobj.__wrapped__
            globalns = getattr(nsobj, "__globals__", None)
        localns = {"unicode": str, **(localns or {})}
        return gth(obj, globalns, localns, include_extras)


if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Literal, TypedDict, Protocol  # noqa F401
else:  # pragma: no cover
    try:
        from typing_extensions import Literal, TypedDict, Protocol  # noqa F401
    except ImportError:
        Literal = NO_TYPE
        TypedDict = NO_TYPE

Annotated = Annotated
Literal = Literal
TypedDict = TypedDict

_T = TypeVar("_T")
_AnnotatedAlias = type(Annotated[_T, ...])
_GenericAlias = type(Generic[_T])
_LiteralMeta = type(Literal)
_TypedDictMeta = type(TypedDict)



from contextlib import suppress
from dataclasses import Field, MISSING, fields, is_dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.types import AnyType
from apischema.typing import get_type_hints

PREFIX = "_apischema_"
Nil = object()


def is_hashable(obj) -> bool:
    try:
        hash(obj)
    except Exception:  # should be TypeError, but who knows what can happen
        return False
    else:
        return True


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


def to_camel_case(s: str):
    pascal_case = "".join(map(str.capitalize, s.split("_")))
    return pascal_case[0].lower() + pascal_case[1:]


def as_dict(obj) -> Dict[str, Any]:
    """like dataclasses.asdict but without deepcopy"""
    assert is_dataclass(obj)
    return {f.name: getattr(obj, f.name) for f in fields(obj)}


_type_hints: Dict[str, Mapping[str, Type]] = {}


def type_hints_cache(obj):
    try:
        return _type_hints[obj]
    except (KeyError, TypeError):
        types = get_type_hints(obj, include_extras=True)
        with suppress(TypeError):
            _type_hints[obj] = types
        return types


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    if field.default is not MISSING:
        return field.default
    raise NotImplementedError()


def type_name(cls: AnyType) -> str:
    for attr in ("__name__", "name", "_name"):
        if hasattr(cls, attr):
            return getattr(cls, attr)
    return str(cls)


MakeDataclassField = Union[Tuple[str, AnyType], Tuple[str, AnyType, Any]]

T = TypeVar("T")


def merge_opts(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    def wrapper(opt1, opt2):
        if opt1 is None:
            return opt2
        if opt2 is None:
            return opt1
        return func(opt1, opt2)

    return wrapper


K = TypeVar("K")
V = TypeVar("V")


@merge_opts
def merge_opts_mapping(m1: Mapping[K, V], m2: Mapping[K, V]) -> Mapping[K, V]:
    return {**m1, **m2}



from collections import defaultdict
from dataclasses import is_dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Collection,
    DefaultDict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    OrderedDict,
    PRIMITIVE_TYPES,
    TUPLE_TYPE,
)
from apischema.typing import Literal, _AnnotatedAlias, _LiteralMeta, _TypedDictMeta
from apischema.utils import type_hints_cache

PRIMITIVE_TYPE_IDS = set(map(id, PRIMITIVE_TYPES))


class Unsupported(TypeError):
    def __init__(self, cls: Type):
        self.cls = cls


Arg = TypeVar("Arg", contravariant=True)
Return = TypeVar("Return", covariant=True)


class Visitor(Generic[Arg, Return]):
    def __init__(self):
        self._generics: DefaultDict[TypeVar, List[AnyType]] = defaultdict(list)

    def annotated(self, cls: AnyType, annotations: Sequence[Any], arg: Arg) -> Return:
        return self.visit(cls, arg)

    def any(self, arg: Arg) -> Return:
        raise NotImplementedError()

    def collection(
        self, cls: Type[Collection], value_type: AnyType, arg: Arg
    ) -> Return:
        raise NotImplementedError()

    def dataclass(self, cls: Type, arg: Arg) -> Return:
        raise NotImplementedError()

    def enum(self, cls: Type[Enum], arg: Arg) -> Return:
        raise NotImplementedError()

    def _generic(self, cls: AnyType, arg: Arg) -> Return:
        origin, args = cls.__origin__, cls.__args__
        assert len(origin.__parameters__) == len(args)
        # Use a side effect in order to avoid passing argument anywhere
        for tv, value in zip(origin.__parameters__, args):
            self._generics[tv].append(value)
        try:
            return self.visit(origin, arg)
        finally:
            for tv in origin.__parameters__:
                self._generics[tv].pop()

    def literal(self, values: Sequence[Any], arg: Arg) -> Return:
        raise NotImplementedError()

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType, arg: Arg
    ) -> Return:
        raise NotImplementedError()

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
        arg: Arg,
    ) -> Return:
        raise NotImplementedError("NamedTuple is not handled")

    def new_type(self, cls: AnyType, super_type: AnyType, arg: Arg) -> Return:
        return self.visit(super_type, arg)

    def primitive(self, cls: Type, arg: Arg) -> Return:
        raise NotImplementedError()

    def subprimitive(self, cls: Type, superclass: Type, arg: Arg) -> Return:
        return self.primitive(superclass, arg)

    def tuple(self, types: Sequence[AnyType], arg: Arg) -> Return:
        raise NotImplementedError()

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool, arg: Arg
    ) -> Return:
        raise NotImplementedError()

    def _type_var(self, tv: AnyType, arg: Arg) -> Return:
        try:
            cls_ = self._generics[tv].pop()
        except IndexError:
            if tv.__constraints__:
                return self.visit(Union[tv.__constraints__], arg)
            else:
                return self.visit(Any, arg)
        try:
            return self.visit(cls_, arg)
        finally:
            self._generics[tv].append(cls_)

    def union(self, alternatives: Sequence[AnyType], arg: Arg) -> Return:
        raise NotImplementedError()

    def _unsupported(self, cls: AnyType, arg: Arg) -> Return:
        if hasattr(cls, "__parameters__") and cls.__parameters__:
            params = tuple(
                (self._generics[p] or [Any])[-1] for p in getattr(cls, "__parameters__")
            )
            if len(params) == 1:
                params = params[0]
            return self.unsupported(cls[params], arg)
        else:
            return self.unsupported(cls, arg)

    def unsupported(self, cls: AnyType, arg: Arg) -> Return:
        raise Unsupported(cls) from None

    def visit(self, cls: AnyType, arg: Arg) -> Return:
        # Use `id` to avoid useless costly generic types hashing
        if id(cls) in PRIMITIVE_TYPE_IDS:
            return self.primitive(cls, arg)
        origin = getattr(cls, "__origin__", None)  # because of 3.6
        if origin is not None:
            if isinstance(cls, _AnnotatedAlias):
                return self.annotated(cls.__args__[0], cls.__metadata__, arg)
            if origin is Union:
                return self.union(cls.__args__, arg)
            if origin is TUPLE_TYPE:
                if len(cls.__args__) < 2 or cls.__args__[1] is not ...:
                    return self.tuple(cls.__args__, arg)
            if origin in COLLECTION_TYPES:
                return self.collection(origin, cls.__args__[0], arg)
            if origin in MAPPING_TYPES:
                return self.mapping(origin, cls.__args__[0], cls.__args__[1], arg)
            if origin is Literal:  # pragma: no cover (because of py36)
                return self.literal(cls.__args__, arg)
            # TypeVar handling
            if not hasattr(origin, "__parameters__"):  # pragma: no cover
                return self.unsupported(origin, arg)
            return self._generic(cls, arg)
        if hasattr(cls, "__supertype__"):
            return self.new_type(cls, cls.__supertype__, arg)
        if isinstance(cls, TypeVar):  # type: ignore
            return self._type_var(cls, arg)
        if cls is Any:
            return self.any(arg)
        # cannot use isinstance(..., TypedDict)
        if isinstance(cls, _TypedDictMeta):
            total = cls.__total__  # type: ignore
            assert isinstance(cls, type)
            return self.typed_dict(cls, type_hints_cache(cls), total, arg)
        return self.visit_not_builtin(cls, arg)

    def visit_not_builtin(self, cls: Type, arg: Arg) -> Return:
        if is_dataclass(cls):
            return self.dataclass(cls, arg)
        # cannot use issubclass before Any or 3.6 Literal
        if isinstance(cls, _LiteralMeta):  # pragma: no cover (because of py36)
            return self.literal(cls.__values__, arg)  # type: ignore
        try:
            issubclass(cls, object)
        except TypeError:
            return self._unsupported(cls, arg)
        if issubclass(cls, Enum):
            return self.enum(cls, arg)
        for primitive in PRIMITIVE_TYPES:
            if issubclass(cls, primitive):
                return self.subprimitive(cls, primitive, arg)
        # NamedTuple
        if issubclass(cls, tuple) and hasattr(cls, "_fields"):
            if hasattr(cls, "__annotations__"):
                types = type_hints_cache(cls)
            elif hasattr(cls, "__field_types"):  # pragma: no cover
                types = cls._field_types  # type: ignore
            else:  # pragma: no cover
                types = OrderedDict((f, Any) for f in cls._fields)  # type: ignore
            return self.named_tuple(
                cls, types, cls._field_defaults, arg  # type: ignore
            )
        return self._unsupported(cls, arg)


class VisitorMock:
    def __init__(self):
        self._method = None
        self._args = None

    def __getattr__(self, name):
        def set_method(*args):
            assert args[-1] is ...
            self._method = name
            self._args = args[:-1]

        return set_method

    def compute_method(
        self, cls: AnyType, visitor_cls: Optional[Type[Visitor]]
    ) -> Callable[[Visitor[Arg, Return], AnyType, Arg], Return]:
        Visitor.visit(cast(Visitor, self), cls, ...)
        method_name, args = self._method, self._args
        assert method_name is not None and args is not None
        if visitor_cls is None:

            def method(visitor: Visitor[Arg, Return], cls: AnyType, arg: Arg) -> Return:
                return getattr(visitor, method_name)(*args, arg)

        else:
            _method = getattr(visitor_cls, method_name)
            if args == (cls,):
                method = _method
            else:

                def method(
                    visitor: Visitor[Arg, Return], cls: AnyType, arg: Arg
                ) -> Return:
                    return _method(visitor, *args, arg)

        return method


def visitor_method(
    cls: AnyType, visitor_cls: Type[Visitor] = None
) -> Callable[[Visitor[Arg, Return], AnyType, Arg], Return]:
    return VisitorMock().compute_method(cls, visitor_cls)



from dataclasses import (  # type: ignore
    Field,
    MISSING,
    _FIELDS,
    _FIELD_CLASSVAR,
    dataclass,
)
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Mapping, Optional, TYPE_CHECKING, Type, TypeVar

from apischema.fields import FIELDS_SET_ATTR
from apischema.utils import get_default

if TYPE_CHECKING:
    from apischema.validation.validator import Validator

MOCK_FIELDS_FIELD = "__mock_fields__"
MOCK_CLS_FIELD = "__mock_cls__"


class NonTrivialDependency(Exception):
    def __init__(self, attr: str):
        self.attr = attr
        self.validator: Optional["Validator"] = None


@dataclass
class ValidatorMock:
    def __init__(self, cls: Type, fields: Mapping[str, Any]):
        self.cls = cls
        self.fields = fields
        setattr(self, FIELDS_SET_ATTR, set(fields))

    def __getattribute__(self, name: str) -> Any:
        fields = super().__getattribute__("fields")
        if name in fields:
            return fields[name]
        cls = super().__getattribute__("cls")
        cls_fields: Mapping[str, Field] = getattr(cls, _FIELDS)
        if name in cls_fields:
            if cls_fields[name]._field_type == _FIELD_CLASSVAR:  # type: ignore
                return getattr(cls, name)
            else:
                try:
                    return get_default(cls_fields[name])
                except NotImplementedError:
                    raise NonTrivialDependency(name)
        if name == "__class__":
            return cls
        if name == "__dict__":
            return {
                **fields,
                **{
                    name: get_default(field)
                    for name, field in cls_fields.items()
                    if field._field_type != _FIELD_CLASSVAR  # type: ignore
                    and (
                        field.default is not MISSING
                        or field.default_factory is not MISSING  # type: ignore
                    )
                },
                FIELDS_SET_ATTR: set(fields),
            }
        if name == FIELDS_SET_ATTR:
            return set(fields)
        if hasattr(cls, name):
            member = getattr(cls, name)
            # for classmethod (staticmethod are not handled)
            if isinstance(member, MethodType):
                return member
            if isinstance(member, FunctionType):
                return partial(member, self)
            if isinstance(member, property):
                return member.fget(self)  # type: ignore
            return member
        raise NonTrivialDependency(name)


T = TypeVar("T")



from enum import Enum
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from apischema import settings
from apischema.conversion.utils import Conversions
from apischema.conversion.visitor import (
    Deserialization,
    DeserializationVisitor,
)
from apischema.dataclasses import is_dataclass
from apischema.dataclasses.cache import (
    Field,
    FieldKind,
    get_deserialization_fields,
    has_post_init_fields,
)
from apischema.deserialization.coercion import Coercer, Coercion, get_coercer
from apischema.json_schema.constraints import (
    ArrayConstraints,
    Constraints,
    get_constraints,
)
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.schema import Schema
from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    LIST_TYPE,
    MAPPING_TYPES,
    NoneType,
    OrderedDict,
    Skip,
    Skipped,
)
from apischema.utils import get_default
from apischema.validation.errors import (
    ErrorKey,
    ErrorMsg,
    ValidationError,
    merge_errors,
)
from apischema.validation.mock import ValidatorMock
from apischema.validation.validator import (
    Validator,
    Validators,
    get_validators,
    validate,
)


def not_one_of(values: Iterable[Any]) -> ValidationError:
    allowed_values = [
        value if not isinstance(value, Enum) else value.value for value in values
    ]
    return ValidationError([f"not one of {allowed_values}"])


def validate_with_errors(
    data: Any,
    constraints: Optional[Constraints],
    errors: Mapping[ErrorKey, ValidationError],
):
    if constraints is not None:
        try:
            constraints.validate(data)
        except ValidationError as err:
            raise ValidationError(err.messages, errors)
    if errors:
        raise ValidationError(children=errors)


T = TypeVar("T")

DataWithConstraint = Tuple[Any, Optional[Constraints]]


class Deserializer(DeserializationVisitor[DataWithConstraint, Any],):
    def __init__(
        self,
        conversions: Optional[Conversions],
        coercer: Coercer,
        additional_properties: bool,
        default_fallback: bool,
    ):
        super().__init__(conversions)
        self.coercer = coercer
        self.additional_properties = additional_properties
        self.default_fallback = default_fallback

    def annotated(
        self, cls: AnyType, annotations: Sequence[Any], data2: DataWithConstraint
    ):
        data, constraints = data2
        validators: Optional[Sequence[Validator]] = None
        # Highest schema is the last annotations
        for annotation in reversed(annotations):
            if annotation is Skip:
                raise Skipped()
            if isinstance(annotation, Schema):
                if constraints is None:
                    constraints = annotation.constraints
                else:
                    constraints = constraints.merge(annotation.constraints)
            if isinstance(annotation, Validators):
                validators = annotation.validators
        result = self.visit(cls, (data, constraints))
        return validate(result, validators) if validators is not None else result

    def any(self, data2: DataWithConstraint):
        return data2[0]

    def collection(
        self, cls: Type[Iterable], value_type: AnyType, data2: DataWithConstraint
    ):
        data, constraints = data2
        data = self.coercer(list, data)
        elts = []
        errors: Dict[ErrorKey, ValidationError] = OrderedDict()
        for i, elt in enumerate(data):
            try:
                elts.append(self.visit(value_type, (elt, None)))
            except ValidationError as err:
                errors[i] = err
        validate_with_errors(data, constraints, errors)
        return elts if cls is LIST_TYPE else COLLECTION_TYPES[cls](elts)

    def dataclass(self, cls: Type, data2: DataWithConstraint):
        assert is_dataclass(cls)
        data, constraints = data2
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        data = self.coercer(dict, data)
        values: Dict[str, Any] = {}
        aliases: List[str] = []
        errors: List[ErrorMsg] = []
        field_errors: Dict[ErrorKey, ValidationError] = OrderedDict()

        def set_field(field: Field, value: Any, alias: Optional[str] = None):
            try:
                result = field.deserialization_method(
                    self, field.deserialization_type, (value, field.constraints)
                )
            except ValidationError as err:
                if field.default and (self.default_fallback or field.default_fallback):
                    pass
                elif alias is not None:
                    field_errors[alias] = err
                else:
                    errors.extend(err.messages)
                    field_errors.update(err.children)
            else:
                values[field.name] = result

        (
            fields,
            merged_fields,
            pattern_fields,
            additional_field,
        ) = get_deserialization_fields(cls)
        for field in fields:
            alias = field.alias
            if alias in data:
                aliases.append(alias)
                set_field(field, data[alias], alias)
            elif field.default:
                if field.deserialization_required_by:
                    requiring = field.deserialization_required_by & data.keys()
                    if requiring:
                        field_errors[alias] = ValidationError(
                            [f"missing property (required by {sorted(requiring)})"]
                        )
            else:
                field_errors[alias] = ValidationError(["missing property"])
        for merged_alias, field in merged_fields:
            merged_alias = merged_alias & data.keys()
            aliases.extend(merged_alias)
            set_field(field, {alias: data[alias] for alias in merged_alias})
        if len(data) != len(aliases):
            remain = data.keys() - set(aliases)
            for pattern, field in pattern_fields:
                if pattern is ...:
                    pattern = infer_pattern(field)
                assert isinstance(pattern, Pattern)
                matched = {key: data[key] for key in remain if pattern.match(key)}
                remain -= matched.keys()
                set_field(field, matched)
            if additional_field is not None:
                set_field(additional_field, {key: data[key] for key in remain})
            elif remain and not self.additional_properties:
                for key in sorted(remain):
                    field_errors[key] = ValidationError(["unexpected property"])
        else:
            for _, field in pattern_fields:
                set_field(field, {})
            if additional_field is not None:
                set_field(additional_field, {})
        error: Optional[ValidationError] = None
        if field_errors or errors:
            error = ValidationError(errors, field_errors)
        if constraints is not None:
            try:
                constraints.validate(data)
            except ValidationError as err:
                error = merge_errors(error, err)
        init: Dict[str, Any] = {}
        if has_post_init_fields(cls):
            for field in fields:
                if field.kind == FieldKind.INIT:
                    if field.name in values:
                        init[field.name] = values[field.name]
                    elif field.name not in field_errors and field.default:
                        init[field.name] = get_default(field.base_field)
        # Don't keep validators when all dependencies are default
        validators = [v for v in get_validators(cls) if v.dependencies & values.keys()]
        if error:
            invalid_fields = field_errors.keys() | {
                field.name for field in fields if field.post_init
            }
            validators = [v for v in validators if not v.dependencies & invalid_fields]
            try:
                validate(ValidatorMock(cls, values), validators, **init)
            except ValidationError as err:
                error = merge_errors(error, err)
            raise error
        try:
            res = cls(**values)
        except (ValidationError, AssertionError):
            raise
        except TypeError as err:
            if str(err).startswith("__init__() got"):
                return self.unsupported(cls, data2)
            else:
                raise ValidationError([str(err)])
        except Exception as err:
            raise ValidationError([str(err)])
        validate(res, validators, **init)
        return res

    def enum(self, cls: Type[Enum], data2: DataWithConstraint):
        data, _ = data2
        assert _ is None
        try:
            return cls(data)
        except ValueError:
            return cls(self.literal([elt.value for elt in cls], data2))

    def literal(self, values: Sequence[Any], data2: DataWithConstraint):
        # Literal can contain Enum values and thus has to call visit too
        data, _ = data2
        assert _ is None
        if data in values:
            return data
        for value in values:
            try:
                if self.visit(type(value), data2) == value:
                    return value
            except Exception:
                continue
        allowed_values = [
            value if not isinstance(value, Enum) else value.value for value in values
        ]
        raise ValidationError([f"not one of {allowed_values}"])

    def mapping(
        self,
        cls: Type[Mapping],
        key_type: AnyType,
        value_type: AnyType,
        data2: DataWithConstraint,
    ):
        data, constraints = data2
        data = self.coercer(dict, data)
        items = {}
        errors: Dict[ErrorKey, ValidationError] = OrderedDict()
        for key, value in data.items():
            assert isinstance(key, str)
            try:
                new_key = self.visit(key_type, (key, None))
                items[new_key] = self.visit(value_type, (value, None))
            except ValidationError as err:
                errors[key] = err
        validate_with_errors(data, constraints, errors)
        return MAPPING_TYPES[cls](items)

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
        data2: DataWithConstraint,
    ):
        data, constraints = data2
        data = self.coercer(dict, data)
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        mapping: Dict[str, Any] = {}
        errors: Dict[ErrorKey, ValidationError] = OrderedDict()
        for key, cls in types.items():
            if key in data:
                try:
                    mapping[key] = self.visit(cls, (data[key], None))
                except ValidationError as err:
                    if key in defaults and self.default_fallback:
                        pass
                    else:
                        errors[key] = err
            elif key in defaults:
                pass
            else:
                errors[key] = ValidationError(["missing property"])
        if not self.additional_properties:
            for key in sorted(data.keys() - defaults.keys() - mapping.keys()):
                errors[key] = ValidationError(["unexpected property"])
        validate_with_errors(data, constraints, errors)
        return validate(cls(**mapping))

    def new_type(self, cls: AnyType, super_type: AnyType, data2: DataWithConstraint):
        data, constraints = data2
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        return validate(
            self.visit(super_type, (data, constraints)), get_validators(cls),
        )

    def primitive(self, cls: Type, data2: DataWithConstraint):
        data, constraints = data2
        try:
            data = self.coercer(cls, data)
        except ValidationError:
            if cls is float and isinstance(data, int):
                data = float(data)
            else:
                raise
        if constraints is not None and data is not None:
            constraints.validate(data)
        return data

    def subprimitive(self, cls: Type, superclass: Type, data2: DataWithConstraint):
        data, constraints = data2
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        return validate(cls(self.primitive(superclass, (data, constraints))))

    def tuple(self, types: Sequence[AnyType], data2: DataWithConstraint):
        data, constraints = data2
        data = self.coercer(list, data)
        ArrayConstraints(min_items=len(types), max_items=len(types)).validate(data)
        elts: List[Any] = []
        errors: Dict[ErrorKey, ValidationError] = OrderedDict()
        for i, (cls, elt) in enumerate(zip(types, data)):
            try:
                elts.append(self.visit(cls, (elt, None)))
            except ValidationError as err:
                errors[i] = err
        validate_with_errors(data, constraints, errors)
        return tuple(elts)

    def typed_dict(
        self,
        cls: Type,
        keys: Mapping[str, AnyType],
        total: bool,
        data2: DataWithConstraint,
    ):
        data, constraints = data2
        data = self.coercer(dict, data)
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        mapping: Dict[str, Any] = {}
        keys_in_dict = 0
        errors: Dict[ErrorKey, ValidationError] = OrderedDict()
        for key, value in data.items():
            assert isinstance(key, str)
            if key in keys:
                keys_in_dict += 1
                try:
                    mapping[key] = self.visit(keys[key], (value, None))
                except ValidationError as err:
                    errors[key] = err
            else:
                mapping[key] = value
        if total and keys_in_dict != len(keys):
            for key in keys:
                if key not in data:
                    errors[key] = ValidationError(["missing property"])
        validate_with_errors(data, constraints, errors)
        return validate(mapping, get_validators(cls))

    def union(self, alternatives: Sequence[AnyType], data2: DataWithConstraint):
        # Optional optimization
        data, _ = data2
        if data is None and alternatives[-1] is NoneType:
            return None
        error: Optional[ValidationError] = None
        for cls in alternatives:
            try:
                return self.visit(cls, data2)
            except Skipped:
                pass
            except ValidationError as err:
                error = merge_errors(error, err)
        raise error if error is not None else RuntimeError("Empty union")

    def visit_conversion(
        self, cls: Type, conversion: Deserialization, data2: DataWithConstraint
    ):
        assert conversion
        data, constraints = data2
        if constraints is not None:
            constraints = constraints.merge(get_constraints(cls))
        else:
            constraints = get_constraints(cls)
        data2 = data, constraints
        error: Optional[ValidationError] = None
        value: Any = ...
        conversions = self.conversions
        for cls_, (converter, self.conversions) in conversion.items():
            try:
                value = self.visit(cls_, data2)
                break
            except Skipped:
                raise TypeError("Deserialization type cannot be skipped")
            except ValidationError as err:
                error = merge_errors(error, err)
            finally:
                self.conversions = conversions
        else:
            assert error is not None
            raise error
        assert value is not ...
        try:
            return validate(converter(value))
        except (ValidationError, AssertionError):
            raise
        except Exception as err:
            raise ValidationError([str(err)])


@overload
def deserialize(
    cls: Type[T],
    data: Any,
    *,
    conversions: Conversions = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
) -> T:
    ...


@overload
def deserialize(
    cls: AnyType,
    data: Any,
    *,
    conversions: Conversions = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
) -> Any:
    ...


def deserialize(
    cls: Type[T],
    data: Any,
    *,
    conversions: Conversions = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
) -> T:
    if additional_properties is None:
        additional_properties = settings.additional_properties
    if coercion is None:
        coercion = settings.coercion
    if default_fallback is None:
        default_fallback = settings.default_fallback
    return Deserializer(
        conversions, get_coercer(coercion), additional_properties, default_fallback
    ).visit(cls, (data, None))



from dataclasses import Field, dataclass, field
from operator import not_
from typing import cast

from pytest import raises

from apischema import ValidationError, validator
from apischema.fields import fields
from apischema.validation.mock import NonTrivialDependency, ValidatorMock
from apischema.validation.validator import Validator, get_validators, validate, validators


@dataclass
class Data:
    a: int
    b: int
    c: int = 0
    with_validator: bool = field(default=False, metadata=validators(not_))

    @validator
    def a_gt_10(self):
        if self.a <= 10:
            yield "error"

    @validator
    def a_lt_100(self):
        if self.a >= 100:
            raise ValueError("error2")

    @validator
    def non_trivial(self):
        non_trivial(self)


validator_field = cast(Field, fields(Data).with_validator)


def non_trivial(data: Data):
    return data.c == data.b


def test_get_validators():
    assert get_validators(Data) == (Data.a_gt_10, Data.a_lt_100, Data.non_trivial)


def test_validator_descriptor():
    # Class field is descriptor
    val: Validator = Data.a_gt_10
    assert val.dependencies == {"a"}
    # Can be called from class and instance
    with raises(ValueError):
        assert Data(200, 0).a_lt_100()
    with raises(ValueError):
        assert Data.a_lt_100(Data(200, 0))


def test_validate():
    validate(Data(42, 0))
    with raises(ValidationError) as err:
        validate(Data(0, 0))
    assert err.value == ValidationError(["error"])
    with raises(ValidationError) as err:
        validate(Data(200, 0))
    assert err.value == ValidationError(["error2"])


def test_non_trivial():
    with raises(NonTrivialDependency) as err:
        validate(ValidatorMock(Data, {"a": 42}))
    # err.value.attr != "c" because `c` has a default value
    assert err.value.attr == "b"
    assert err.value.validator == Data.non_trivial



import collections.abc
import sys
from types import MappingProxyType
from typing import (
    AbstractSet,
    Any,
    Collection,
    Dict,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.typing import Annotated, NO_TYPE

AnyType = Any
NoneType: Type[None] = type(None)
Number = Union[int, float]

PRIMITIVE_TYPES = (str, int, bool, float, NoneType)

if sys.version_info >= (3, 7):  # pragma: no cover
    COLLECTION_TYPES = {
        collections.abc.Collection: tuple,
        collections.abc.Sequence: tuple,
        tuple: tuple,
        collections.abc.MutableSequence: list,
        list: list,
        collections.abc.Set: frozenset,
        frozenset: frozenset,
        collections.abc.MutableSet: set,
        set: set,
    }
    MAPPING_TYPES = {
        collections.abc.Mapping: MappingProxyType,
        collections.abc.MutableMapping: dict,
        dict: dict,
    }
    LIST_TYPE = list
    TUPLE_TYPE = tuple
    DICT_TYPE = dict

else:  # pragma: no cover
    COLLECTION_TYPES = {
        Collection: tuple,
        Sequence: tuple,
        Tuple: tuple,
        MutableSequence: list,
        List: list,
        AbstractSet: frozenset,
        FrozenSet: frozenset,
        Set: set,
    }
    MAPPING_TYPES = {Mapping: MappingProxyType, MutableMapping: dict, Dict: dict}
    LIST_TYPE = List
    TUPLE_TYPE = Tuple
    DICT_TYPE = Dict


if sys.version_info >= (3, 7):  # pragma: no cover
    OrderedDict = dict
else:  # pragma: no cover
    from collections import OrderedDict  # noqa

# Kind of hack to benefit of PEP 584
if sys.version_info >= (3, 9):  # pragma: no cover
    Metadata = Mapping[str, Any]
else:  # pragma: no cover

    class Metadata(Mapping[str, Any]):
        def __or__(self, other: Mapping[str, Any]) -> "Metadata":
            return MappingWithUnion({**self, **other})

        def __ror__(self, other: Mapping[str, Any]) -> "Metadata":
            return MappingWithUnion({**other, **self})


class MetadataMixin(Metadata):
    def __init__(self, key: str):
        super().__setattr__("_key", key)

    def __getitem__(self, key):
        if key != self._key:
            raise KeyError(key)
        return self

    def __iter__(self):
        return iter((self._key,))

    def __len__(self):
        return 1


if sys.version_info >= (3, 9):
    MappingWithUnion = MappingProxyType
else:

    class MappingWithUnion(dict, Metadata):
        pass


class Skipped(Exception):
    pass


class _Skip:
    def __call__(self, *, schema_only: bool):
        return SkipSchema if schema_only else self


Skip = _Skip()
SkipSchema = object()


T = TypeVar("T")

if Annotated is not NO_TYPE:
    NotNull = Union[T, Annotated[None, Skip]]
else:

    class _NotNull:
        def __getitem__(self, item):
            raise TypeError("NotNull requires Annotated (PEP 593)")

    NotNull = _NotNull()  # type: ignore



from collections import ChainMap
from copy import copy
from dataclasses import Field, field, make_dataclass
from inspect import getmembers
from itertools import chain
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema import schema_ref, serializer
from apischema.conversion import conversions
from apischema.conversion.converters import extra_serializer
from apischema.conversion.utils import Conversions
from apischema.dataclasses.cache import get_aggregate_serialization_fields
from apischema.fields import with_fields_set
from apischema.typing import get_type_hints
from apischema.utils import MakeDataclassField

Metadata = Mapping[str, Any]


def _check_field(field: Any) -> str:
    if isinstance(field, Field):
        return field.name
    elif isinstance(field, str):
        return field
    elif isinstance(field, property):
        return field.fget.__name__
    else:
        raise TypeError("Serialization fields must be Field or str or property")


Cls = TypeVar("Cls", bound=Type)
FieldFactory = Callable[[Cls], Any]


def dataclass_serializer(
    cls: Cls,
    ref: Optional[str] = None,
    *,
    include: Optional[Collection[Any]] = None,
    include_properties: bool = False,
    exclude: Collection[str] = (),
    override: Optional[Mapping[str, Optional[Metadata]]] = None,
    additional: Optional[
        Mapping[str, Union[FieldFactory, Tuple[FieldFactory, Conversions]]]
    ] = None,
    as_default_serializer: bool = False,
) -> Type:
    if override is None:
        override = {}
    override = {_check_field(field_): metadata for field_, metadata in override.items()}
    if include is not None:
        include = set(map(_check_field, include)) | override.keys()
    exclude = set(map(_check_field, exclude))
    fields, properties_fields = get_aggregate_serialization_fields(cls)
    new_fields: Dict[str, MakeDataclassField] = {}
    for field_ in chain(fields, properties_fields):
        if field_.name in exclude or (
            include is not None and field_.name not in include
        ):
            continue
        new_field = copy(field_.base_field)
        metadata = override.get(field_.name)
        if metadata is not None:
            new_field.metadata = ChainMap(metadata, new_field.metadata)
        new_fields[field_.name] = field_.name, field_.type, new_field
    for name, prop in getmembers(cls, lambda m: isinstance(m, property)):
        if name in exclude or (
            not include_properties and include is not None and name not in include
        ):
            continue
        assert isinstance(prop, property)
        types = get_type_hints(prop.fget, include_extras=True)
        try:
            new_fields[name] = name, types["return"]
        except KeyError:
            raise TypeError("Properties serialized must be typed") from None
    attributes_set = set(new_fields)
    factories: Dict[str, FieldFactory] = {}
    for name, factory in (additional or {}).items():
        attributes_set.discard(name)
        conv: Optional[Conversions]
        if isinstance(factory, tuple):
            factory, conv = factory
        else:
            conv = None
        factories[name] = factory
        new_field = field(metadata=conversions(serialization=conv))
        types = get_type_hints(factory, include_extras=True)
        try:
            new_fields[name] = name, types["return"], new_field
        except KeyError:
            raise TypeError("Additional field factories must be typed") from None
    attributes_list = list(attributes_set)  # for iteration performance
    serialized_class = with_fields_set(
        make_dataclass(f"{cls.__name__}Serializer", new_fields.values())
    )
    if ref is ...:
        raise TypeError("Dataclass serializer ref cannot be ...")
    schema_ref(ref)(serialized_class)
    register = serializer if as_default_serializer else extra_serializer

    def converter(obj):
        return serialized_class(
            **{name: getattr(obj, name) for name in attributes_list},
            **{name: factory(obj) for name, factory in factories.items()},
        )

    register(converter, cls, serialized_class)
    return serialized_class



from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Sequence

from apischema.utils import Nil


@dataclass(frozen=True)
class Annotations:
    extra: Optional[Mapping[str, Any]] = None
    extra_only: bool = False
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = Nil
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None


_annotations: Dict[Any, Annotations] = {}

get_annotations = _annotations.get

NO_DEFAULT = object()



from typing import AbstractSet, Any, Sequence, Type

from apischema.conversion.visitor import (
    Conv,
    ConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.dataclasses.cache import (
    Field,
    FieldCache,
    get_deserialization_fields,
    get_serialization_fields,
)
from apischema.fields import fields
from apischema.types import AnyType, Skip, SkipSchema, Skipped
from apischema.visitor import Arg, Return


class SchemaVisitor(ConversionsVisitor[Conv, Arg, Return]):
    def annotated(self, cls: AnyType, annotations: Sequence[Any], arg: Arg) -> Return:
        for annotation in reversed(annotations):
            if annotation in {Skip, SkipSchema}:
                raise Skipped
        return self._annotated(cls, annotations, arg)

    def _annotated(self, cls: AnyType, annotations: Sequence[Any], arg: Arg) -> Return:
        return self.visit(cls, arg)

    @staticmethod
    def _dataclass_fields(cls: Type) -> FieldCache:
        raise NotImplementedError()

    _field_type: Any
    _field_conversions: Any

    def _field_visit(self, field: Field, arg: Arg) -> Return:
        conversions_save = self.conversions
        self.conversions = getattr(field, self._field_conversions.name)
        try:
            return self.visit(getattr(field, self._field_type.name), arg)
        finally:
            self.conversions = conversions_save

    _field_required_by: Any

    @classmethod
    def _required_by(cls, field: Field) -> AbstractSet[str]:
        return getattr(field, cls._field_required_by.name)

    @staticmethod
    def _override_arg(cls: AnyType, arg: Arg) -> Arg:  # type: ignore
        return arg

    def _union_arg(self, cls: AnyType, arg: Arg) -> Arg:  # type: ignore
        raise NotImplementedError()

    def _union_result(self, results: Sequence[Return], arg: Arg) -> Return:
        raise NotImplementedError()

    def union(self, alternatives: Sequence[AnyType], arg: Arg) -> Return:
        results = []
        for cls in alternatives:
            try:
                results.append(self.visit(cls, self._union_arg(cls, arg)))
            except Skipped:
                pass
        return self._union_result(results, arg)

    def _is_conversion(self, cls: AnyType) -> bool:
        # In 3.6, GenericAlias are classes with mro
        if getattr(cls, "__origin__", None) is not None:
            return False
        try:
            return bool(self.is_conversion(cls, self.conversions))
        except Exception:
            return False


class DeserializationSchemaVisitor(
    DeserializationVisitor[Arg, Return], SchemaVisitor[Deserialization, Arg, Return],
):
    def visit_conversion(
        self, cls: Type, conversion: Deserialization, arg: Arg
    ) -> Return:
        assert conversion
        results = []
        conversions = self.conversions
        for cls_, (converter, self.conversions) in conversion.items():
            try:
                results.append(self.visit(cls_, self._union_arg(cls, arg)))
            except Skipped:
                raise TypeError("Deserialization type cannot be skipped")
            finally:
                self.conversions = conversions
        return self._union_result(results, self._override_arg(cls, arg))

    _dataclass_fields = staticmethod(get_deserialization_fields)  # type: ignore
    _field_type = fields(Field).deserialization_type
    _field_conversions = fields(Field).deserialization_conversions
    _field_required_by = fields(Field).deserialization_required_by


class SerializationSchemaVisitor(
    SerializationVisitor[Arg, Return], SchemaVisitor[Serialization, Arg, Return],
):
    def visit_conversion(
        self, cls: Type, conversion: Serialization, arg: Arg
    ) -> Return:
        conversions = self.conversions
        target, (converter, self.conversions) = conversion
        try:
            return self.visit(target, self._override_arg(cls, arg))
        finally:
            self.conversions = conversions

    _dataclass_fields = staticmethod(get_serialization_fields)  # type: ignore
    _field_type = fields(Field).serialization_type
    _field_conversions = fields(Field).serialization_conversions
    _field_required_by = fields(Field).serialization_required_by



from dataclasses import Field as BaseField
from typing import Dict, Pattern

from apischema.dataclasses.cache import Field

_patterns: Dict[BaseField, Pattern] = {}


def infer_pattern(field: Field) -> Pattern:
    try:
        return _patterns[field.base_field]
    except KeyError:
        from apischema.json_schema.generation.builder import (
            DeserializationSchemaBuilder,
            SerializationSchemaBuilder,
        )

        for schema_builder, cls in (
            (DeserializationSchemaBuilder, field.deserialization_type),
            (SerializationSchemaBuilder, field.serialization_type),
        ):
            try:
                prop_schema = schema_builder(None, lambda s: s, {}, False).visit(cls)
                if (
                    len(prop_schema.get("patternProperties", {})) != 1
                    or "additionalProperties" in prop_schema
                ):
                    continue
                _patterns[field.base_field] = pattern = next(
                    iter(prop_schema["patternProperties"])
                )
                return pattern
            except RecursionError:
                pass
        else:
            raise TypeError("Cannot inferred pattern from type schema") from None



from typing import Generic, TypeVar

from pytest import raises

from apischema.conversion.utils import substitute_type_vars

T = TypeVar("T")
U = TypeVar("U")


class A(Generic[T]):
    pass


class B(Generic[T, U]):
    pass


def test_substitute_type_vars():
    assert substitute_type_vars(A[U], B[U, int]) == (A, B[T, int])
    with raises(TypeError):
        substitute_type_vars(A[int], B[int, int])



from collections import ChainMap
from dataclasses import dataclass, fields
from types import MappingProxyType
from typing import (
    Callable,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.conversion.raw import to_raw_deserializer
from apischema.conversion.utils import Conversions, Converter
from apischema.dataclasses import is_dataclass
from apischema.types import AnyType, MetadataMixin

Cls = TypeVar("Cls", bound=Type)


@dataclass
class ConversionsMetadata(MetadataMixin):
    both: Optional[AnyType] = None
    deserialization: Optional[Conversions] = None
    serialization: Optional[Conversions] = None
    deserializer: Optional[Converter] = None
    serializer: Optional[Converter] = None

    def __post_init__(self):
        from apischema.metadata.keys import CONVERSIONS_METADATA

        super().__init__(CONVERSIONS_METADATA)
        if (
            self.both is not None
            and len([f for f in fields(self) if getattr(self, f.name) is not None]) > 1
        ):
            raise ValueError("Cannot set both and (de)serialization/(de)serializers")

    def __call__(self, dataclass: Cls) -> Cls:
        from apischema.metadata.keys import CONVERSIONS_METADATA, MERGED_METADATA

        if not is_dataclass(dataclass):
            raise TypeError("Must be applied to dataclass")
        for conv in (self.deserialization, self.serialization):
            if conv is not None and not isinstance(conv, Mapping):
                raise TypeError("Dataclasses conversions must be a Mapping instance")
        for field in fields(dataclass):
            if MERGED_METADATA in field.metadata:
                continue
            if CONVERSIONS_METADATA in field.metadata:
                conversions: ConversionsMetadata = field.metadata[CONVERSIONS_METADATA]
                if any((self.deserialization, conversions.deserialization)):
                    conversions.deserialization = ChainMap(
                        conversions.deserialization or {}, self.deserialization or {}
                    )
                if any((self.serialization, conversions.serialization)):
                    conversions.serialization = ChainMap(
                        conversions.serialization or {}, self.serialization or {}
                    )
            else:
                conversions = ConversionsMetadata(
                    deserialization=self.deserialization,
                    serialization=self.serialization,
                )
                field.metadata = MappingProxyType({CONVERSIONS_METADATA: conversions})
        return dataclass


Conversions_ = Union[Mapping[AnyType, AnyType], AnyType]


@overload
def conversions(both: AnyType = None) -> ConversionsMetadata:
    ...


@overload
def conversions(
    *,
    deserialization: Conversions = None,
    serialization: Conversions = None,
    deserializer: Converter = None,
    serializer: Converter = None,
    raw_deserializer: Callable = None,
) -> ConversionsMetadata:
    ...


def conversions(
    both: AnyType = None,
    *,
    deserialization: Conversions = None,
    serialization: Conversions = None,
    deserializer: Converter = None,
    serializer: Converter = None,
    raw_deserializer: Callable = None,
) -> ConversionsMetadata:
    if raw_deserializer is not None:
        deserializer = to_raw_deserializer(raw_deserializer)
    return ConversionsMetadata(
        both, deserialization, serialization, deserializer, serializer
    )



from dataclasses import Field

from apischema.utils import PREFIX

ALIAS_METADATA = f"{PREFIX}alias"
ALIAS_NO_OVERRIDE_METADATA = f"{PREFIX}alias_no_override"
MERGED_METADATA = f"{PREFIX}merged"
CONVERSIONS_METADATA = f"{PREFIX}conversions"
DEFAULT_AS_SET = f"{PREFIX}default_as_set"
DEFAULT_FALLBACK_METADATA = f"{PREFIX}default_fallback"
INIT_VAR_METADATA = f"{PREFIX}init_var"
POST_INIT_METADATA = f"{PREFIX}post_init"
PROPERTIES_METADATA = f"{PREFIX}properties"
REQUIRED_METADATA = f"{PREFIX}required"
SCHEMA_METADATA = f"{PREFIX}schema"
SKIP_METADATA = f"{PREFIX}skip"
VALIDATORS_METADATA = f"{PREFIX}validators"

INCOMPATIBLE_WITH_AGGREGATE = {
    ALIAS_METADATA,
    ALIAS_NO_OVERRIDE_METADATA,
    CONVERSIONS_METADATA,
    REQUIRED_METADATA,
    SCHEMA_METADATA,
    VALIDATORS_METADATA,
}
INCOMPATIBLE_WITH_MERGED = {PROPERTIES_METADATA, *INCOMPATIBLE_WITH_AGGREGATE}
INCOMPATIBLE_WITH_PROPERTIES = {MERGED_METADATA, *INCOMPATIBLE_WITH_AGGREGATE}


def is_aggregate_field(field: Field) -> bool:
    return MERGED_METADATA in field.metadata or PROPERTIES_METADATA in field.metadata



from pytest import raises

from apischema.deserialization import unflat_key_value


def test_items_to_data():
    lines = {
        "key1.0": "v0",
        "key1.2": "v2",
        "key2": 42,
    }
    assert unflat_key_value(lines.items()) == {
        "key1": ["v0", None, "v2"],
        "key2": 42,
    }
    with raises(ValueError):
        unflat_key_value({"": ...}.items())
    with raises(ValueError):
        unflat_key_value({**lines, "key1.key3": ...}.items())
    with raises(ValueError):
        unflat_key_value({**lines, "0": ...}.items())



import sys
from inspect import Parameter, isclass, signature
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from apischema.types import AnyType
from apischema.typing import _GenericAlias, get_type_hints
from apischema.utils import type_name
from apischema.visitor import Visitor

Conversions = Mapping[AnyType, Any]
Converter = Callable[[Any], Any]
ConverterWithConversions = Tuple[Converter, Optional[Conversions]]

Param = TypeVar("Param")
Return = TypeVar("Return")


def check_converter(
    converter: Converter,
    param: Optional[Type[Param]],
    ret: Optional[Type[Return]],
    namespace: Dict[str, Any] = None,
) -> Tuple[Type[Param], Type[Return]]:
    try:
        parameters = iter(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if ret is None and isclass(converter):
            ret = cast(Type[Any], converter)
        if param is None:
            raise TypeError("converter parameter must be typed")
    else:
        try:
            first = next(parameters)
        except StopIteration:
            raise TypeError("converter must have at least one parameter")
        types = get_type_hints(converter, None, namespace, include_extras=True)
        for p in parameters:
            if p.default is Parameter.empty and p.kind not in (
                Parameter.VAR_POSITIONAL,
                Parameter.VAR_KEYWORD,
            ):
                raise TypeError(
                    "converter must have at most one parameter " "without default"
                )
        if param is None:
            try:
                param = types.pop(first.name)
            except KeyError:
                raise TypeError("converter parameter must be typed")
        if ret is None:
            try:
                ret = types.pop("return")
            except KeyError:
                if isclass(converter):
                    ret = cast(Type, converter)
                else:
                    raise TypeError("converter return must be typed")
    return cast(Type[Param], param), cast(Type[Return], ret)


Cls = TypeVar("Cls", bound=Type)
Other = TypeVar("Other", bound=Type)

if sys.version_info >= (3, 7):  # pragma: no cover
    import collections.abc
    import re
    from typing import (  # noqa F811
        List,
        AbstractSet,
        Set,
        Dict,
        Collection,
        Sequence,
        MutableSequence,
        Pattern,
    )

    TYPED_ORIGINS = {
        tuple: Tuple,
        list: List,
        frozenset: AbstractSet,
        set: Set,
        dict: Dict,
        collections.abc.Collection: Collection,
        collections.abc.Sequence: Sequence,
        collections.abc.MutableSequence: MutableSequence,
        collections.abc.Set: AbstractSet,
        collections.abc.MutableSet: Set,
        re.Pattern: Pattern,
    }

    def _get_origin(cls: _GenericAlias) -> Type:  # type: ignore
        return TYPED_ORIGINS.get(cls.__origin__, cls.__origin__)  # type: ignore


else:  # pragma: no cover

    def _get_origin(cls: _GenericAlias) -> Type:  # type: ignore
        return cls.__origin__  # type: ignore


def substitute_type_vars(base: Cls, other: Other) -> Tuple[Cls, Other]:
    if getattr(base, "__origin__", None) is None:
        return base, other
    if not all(isinstance(arg, TypeVar) for arg in base.__args__):  # type: ignore
        raise TypeError(
            f"Generic conversion doesn't support partial specialization,"
            f" aka {type_name(base)}[{','.join(map(type_name, base.__args__))}]"
        )
    substitution = dict(zip(base.__args__, _get_origin(base).__parameters__))
    if isinstance(other, TypeVar):  # type: ignore
        new_other = substitution.get(other, other)
    elif getattr(other, "__origin__", None) is not None:
        new_other = _get_origin(other)[
            tuple(substitution.get(arg, arg) for arg in other.__args__)
        ]
    else:
        new_other = other
    return cast(Tuple[Cls, Other], (_get_origin(base), new_other))


class ConvertibleVisitor(Visitor):
    def annotated(self, cls: AnyType, annotations: Sequence[Any], _):
        raise NotImplementedError()

    def new_type(self, cls: AnyType, super_type: AnyType, _):
        raise NotImplementedError()

    def _type_var(self, tv: AnyType, _):
        raise NotImplementedError()

    def visit_not_builtin(self, cls: AnyType, _):
        return


def is_convertible(cls: AnyType):
    try:
        ConvertibleVisitor().visit(cls, ...)
    except NotImplementedError:
        return False
    else:
        return True


def check_convertible(cls: AnyType):
    if not is_convertible(cls):
        raise TypeError(f"{type_name(cls)} is not a class")



import re
import sys
from base64 import b64decode, b64encode
from datetime import date, datetime, time
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import Path
from typing import NewType, Type, TypeVar
from uuid import UUID

from apischema.json_schema.schema import schema
from apischema.conversion.converters import deserializer, serializer

Cls = TypeVar("Cls", bound=Type)


def as_str(cls: Cls, format: str = None):
    str_type = schema(format=format)(NewType(cls.__name__, str))
    deserializer(cls, str_type, cls)
    serializer(str, cls, str_type)


# =================== bytes =====================

deserializer(b64decode, str, bytes)


@serializer
def to_base64(b: bytes) -> str:
    return b64encode(b).decode()


# ================== datetime ===================

if sys.version_info >= (3, 7):  # pragma: no cover
    for cls, format in [(date, "date"), (datetime, "date-time"), (time, "time")]:
        str_type = schema(format=format)(NewType(cls.__name__, str))
        deserializer(cls.fromisoformat, str_type, cls)  # type: ignore
        serializer(cls.isoformat, cls, str_type)  # type: ignore

# ================== decimal ====================

deserializer(Decimal, float, Decimal)
serializer(float, Decimal, float)

# ================= ipaddress ===================

for cls in (IPv4Address, IPv4Interface, IPv4Network):
    as_str(cls, "ipv4")
for cls in (IPv6Address, IPv6Interface, IPv6Network):
    as_str(cls, "ipv6")

# ==================== path =====================

as_str(Path)

# =================== pattern ===================

Pattern = type(re.compile(r""))
deserializer(re.compile, str, Pattern)
serializer(lambda p: p.pattern, Pattern, str)

# ==================== uuid =====================

as_str(UUID, "uuid")



__all__ = ["serialize"]
from dataclasses import is_dataclass
from enum import Enum
from typing import Any, Collection, Mapping, Optional, Type, TypeVar

from apischema.conversion.utils import Conversions
from apischema.conversion.visitor import (
    Serialization,
    SerializationVisitor,
)
from apischema.dataclasses.cache import get_aggregate_serialization_fields
from apischema.fields import FIELDS_SET_ATTR, fields_set
from apischema.types import AnyType, PRIMITIVE_TYPES

PRIMITIVE_TYPES_SET = set(PRIMITIVE_TYPES)
COLLECTION_TYPE_SET = {list, tuple, set, frozenset}


T = TypeVar("T", bound=Any)


class Serializer(SerializationVisitor[Any, Any]):
    def __init__(self, conversions: Optional[Conversions], exclude_unset: bool):
        super().__init__(conversions)
        self.exclude_unset = exclude_unset

    def visit(self, cls: AnyType, obj):
        return self.visit2(obj)

    def visit2(self, obj):
        cls = type(obj)
        # inline to be faster
        if cls in PRIMITIVE_TYPES_SET:
            return obj
        if cls in COLLECTION_TYPE_SET:
            return [self.visit2(elt) for elt in obj]
        if cls is dict:
            return {self.visit2(key): self.visit2(value) for key, value in obj.items()}
        return self.visit_not_builtin(cls, obj)

    def visit_conversion(self, _, conversion: Serialization, obj):
        conversions = self.conversions
        _, (converter, self.conversions) = conversion
        try:
            return self.visit2(converter(obj))
        finally:
            self.conversions = conversions

    def visit_not_conversion(self, cls: Type, obj):
        assert cls is type(obj)
        if is_dataclass(cls):
            fields, aggregate_fields = get_aggregate_serialization_fields(cls)
            if self.exclude_unset and hasattr(obj, FIELDS_SET_ATTR):
                fields_set_ = fields_set(obj)
                fields = [f for f in fields if f.name in fields_set_]
                aggregate_fields = [
                    f for f in aggregate_fields if f.name in fields_set_
                ]
            result = {}
            # properties before normal fields to avoid overloading a field with property
            for field in aggregate_fields:
                result.update(
                    field.serialization_method(self, getattr(obj, field.name))
                )
            for field in fields:
                result[field.alias] = field.serialization_method(
                    self, getattr(obj, field.name)
                )
            return result
        if issubclass(cls, Enum):
            return obj.value
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, Mapping):
            return {self.visit2(key): self.visit2(value) for key, value in obj.items()}
        if isinstance(obj, Collection):
            return [self.visit2(elt) for elt in obj]
        if issubclass(cls, tuple) and hasattr(cls, "_fields"):
            return {f: self.visit2(getattr(obj, f)) for f in obj._fields}
        return self.unsupported(cls, obj)


NO_OBJ = object()


def serialize(
    obj: Any, *, conversions: Conversions = None, exclude_unset: bool = True,
) -> Any:
    return Serializer(conversions, exclude_unset).visit(type(obj), obj)
